# ruff: noqa: F401


from .v2_dataclasses import UniswapV2PoolSimulationResult, UniswapV2PoolState
from .v2_liquidity_pool import LiquidityPool
from .v3_dataclasses import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
)
from .v3_liquidity_pool import V3LiquidityPool
from .v3_snapshot import UniswapV3LiquiditySnapshot
from .v3_tick_lens import TickLens



from ...constants import MAX_UINT256, MIN_UINT256
from ...exceptions import EVMRevertError
from .functions import mulmod


def mulDiv(
    a: int,
    b: int,
    denominator: int,
):
    """
    The Solidity implementation is designed to calculate a * b / d without risk of overflowing
    the intermediate result (maximum of 2**256-1).

    Python does not have this bit depth limitations on integers,
    so simply check for exceptional conditions then return the result
    """

    if not (MIN_UINT256 <= a <= MAX_UINT256):
        raise EVMRevertError(f"Invalid input, {a} does not fit into uint256")

    if not (MIN_UINT256 <= b <= MAX_UINT256):
        raise EVMRevertError(f"Invalid input, {b} does not fit into uint256")

    if denominator == 0:
        raise EVMRevertError("DIVISION BY ZERO")

    result = (a * b) // denominator

    if not (MIN_UINT256 <= result <= MAX_UINT256):
        raise EVMRevertError("invalid result, will not fit in uint256")

    return result


def mulDivRoundingUp(a: int, b: int, denominator: int):
    result: int = mulDiv(a, b, denominator)
    if mulmod(a, b, denominator) > 0:
        # must be less than max uint256 since we're rounding up
        if not (MIN_UINT256 <= result < MAX_UINT256):
            raise EVMRevertError("FAIL!")
        result += 1
    return result



# ruff: noqa: F401


from .flash_borrow_to_lp_swap import FlashBorrowToLpSwap
from .flash_borrow_to_lp_swap_new import FlashBorrowToLpSwapNew
from .flash_borrow_to_lp_swap_with_future import FlashBorrowToLpSwapWithFuture
from .flash_borrow_to_router_swap import FlashBorrowToRouterSwap
from .uniswap_lp_cycle import UniswapLpCycle, ArbitrageCalculationResult



from ...exceptions import EVMRevertError
from .functions import uint128


def addDelta(x: int, y: int) -> int:
    """
    This function has been heavily modified to directly check that the result
    fits in a uint128, instead of checking via < or >= tricks via Solidity's
    built-in casting as implemented at https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/LiquidityMath.sol
    """

    if not (0 <= x <= 2**128 - 1):
        raise EVMRevertError("x not a valid uint128")

    if not (-(2**127) <= y <= 2**127 - 1):
        raise EVMRevertError("y not a valid int128")

    if y < 0:
        z = x - uint128(-y)
        if not (0 <= z <= 2**128 - 1):
            raise EVMRevertError("LS")

    else:
        z = x + uint128(y)
        if not (0 <= z <= 2**128 - 1):
            raise EVMRevertError("LA")

    return z



import asyncio
from fractions import Fraction
from threading import Lock
from typing import (
    TYPE_CHECKING,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

if TYPE_CHECKING:
    from concurrent.futures import Executor
from warnings import warn

import eth_abi
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from scipy.optimize import minimize_scalar  # type: ignore[import]
from web3 import Web3

from ..exceptions import (
    ArbitrageError,
    EVMRevertError,
    LiquidityPoolError,
    ZeroLiquidityError,
)
from ..logging import logger
from ..erc20_token import Erc20Token
from ..baseclasses import ArbitrageHelper
from ..uniswap.v2_liquidity_pool import CamelotLiquidityPool, LiquidityPool
from ..uniswap.v2_dataclasses import (
    UniswapV2PoolSimulationResult,
    UniswapV2PoolState,
)
from ..uniswap.v3_libraries import TickMath
from ..uniswap.v3_dataclasses import (
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
)
from ..uniswap.v3_liquidity_pool import V3LiquidityPool
from .arbitrage_dataclasses import (
    ArbitrageCalculationResult,
    UniswapPoolSwapVector,
    UniswapV2PoolSwapAmounts,
    UniswapV3PoolSwapAmounts,
)


class UniswapLpCycle(ArbitrageHelper):
    __slots__ = (
        "_lock",
        "_swap_vectors",
        "best",
        "id",
        "input_token",
        "gas_estimate",
        "max_input",
        "name",
        "pool_states",
        "swap_pools",
    )

    def __init__(
        self,
        input_token: Erc20Token,
        swap_pools: Iterable[Union[LiquidityPool, V3LiquidityPool]],
        id: str,
        max_input: Optional[int] = None,
    ):
        self._lock = Lock()

        for pool in swap_pools:
            if not isinstance(pool, (LiquidityPool, V3LiquidityPool)):
                raise ValueError(f"Could not identify Uniswap version for pool {pool}!")
        self.swap_pools = tuple(swap_pools)

        self.id = id
        self.input_token = input_token

        if max_input is None:
            warn("No maximum input provided, setting to 100 WETH")
            max_input = 100 * 10**18
        self.max_input = max_input

        self.gas_estimate: Optional[int] = None

        # self.swap_pool_addresses = [pool.address for pool in self.swap_pools]
        # self.swap_pool_tokens = [
        # [pool.token0, pool.token1] for pool in self.swap_pools
        # ]

        # Set up pre-determined "swap vectors", which allows the helper
        # to identify the tokens and direction of each swap along the path
        _swap_vectors: List[UniswapPoolSwapVector] = []
        for i, pool in enumerate(self.swap_pools):
            if i == 0:
                if self.input_token == pool.token0:
                    token_in = pool.token0
                    token_out = pool.token1
                    zero_for_one = True
                elif self.input_token == pool.token1:
                    token_in = pool.token1
                    token_out = pool.token0
                    zero_for_one = False
                else:
                    raise ValueError("Input token could not be identified!")
            else:
                # token_out references the output from the previous pool
                if token_out == pool.token0:
                    token_in = pool.token0
                    token_out = pool.token1
                    zero_for_one = True
                elif token_out == pool.token1:
                    token_in = pool.token1
                    token_out = pool.token0
                    zero_for_one = False
                else:
                    raise ValueError("Input token could not be identified!")
            _swap_vectors.append(
                UniswapPoolSwapVector(
                    token_in=token_in,
                    token_out=token_out,
                    zero_for_one=zero_for_one,
                )
            )
        self._swap_vectors = tuple(_swap_vectors)

        self.name = " -> ".join([pool.name for pool in self.swap_pools])

        self.pool_states: Dict[
            ChecksumAddress,
            Optional[
                Union[
                    UniswapV2PoolState,
                    UniswapV3PoolState,
                ]
            ],
        ] = {pool.address: None for pool in self.swap_pools}

        self.best: dict = {
            "input_token": self.input_token,
            "last_swap_amount": 0,
            "profit_amount": 0,
            "profit_token": self.input_token,
            "strategy": "cycle",
            "swap_amount": 0,
            "swap_pools": self.swap_pools,
            # "swap_pool_addresses": self.swap_pool_addresses,
            "swap_pool_amounts": [],
            # "swap_pool_tokens": self.swap_pool_tokens,
        }

    def __getstate__(self) -> dict:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        dropped_attributes = ("_lock",)

        with self._lock:
            try:
                self.__slots__
            except AttributeError:
                return {
                    attribute: value
                    for attribute, value in self.__dict__.items()
                    if attribute not in dropped_attributes
                }
            else:
                return {
                    attribute: getattr(self, attribute, None)
                    for attribute in self.__slots__
                    if attribute not in dropped_attributes
                }

    def __setstate__(self, state: dict):
        for key, value in state.items():
            setattr(self, key, value)

    def __str__(self) -> str:
        return self.name

    def _sort_overrides(
        self,
        overrides: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ],
    ) -> Dict[ChecksumAddress, Union[UniswapV2PoolState, UniswapV3PoolState]]:
        """
        Validate the overrides, extract and insert the resulting pool states
        into a dictionary.
        """

        if overrides is None:
            return {}

        sorted_overrides: Dict[ChecksumAddress, Union[UniswapV2PoolState, UniswapV3PoolState]] = {}

        for pool, override in overrides:
            if isinstance(
                override,
                (
                    UniswapV2PoolState,
                    UniswapV3PoolState,
                ),
            ):
                logger.debug(f"Applying override {override} to {pool}")
                sorted_overrides[pool.address] = override
            elif isinstance(
                override,
                (
                    UniswapV2PoolSimulationResult,
                    UniswapV3PoolSimulationResult,
                ),
            ):
                logger.debug(f"Applying override {override.future_state} to {pool}")
                sorted_overrides[pool.address] = override.future_state
            else:
                raise ValueError(f"Override for {pool} has unsupported type {type(override)}")

        return sorted_overrides

    def _build_amounts_out(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        pool_state_overrides: Optional[
            Dict[ChecksumAddress, Union[UniswapV2PoolState, UniswapV3PoolState]]
        ] = None,
    ) -> List[Union[UniswapV2PoolSwapAmounts, UniswapV3PoolSwapAmounts]]:
        """
        Generate human-readable inputs for a complete swap along the arbitrage
        path, starting with `token_in_quantity` amount of `token_in`.
        """

        if pool_state_overrides is None:
            pool_state_overrides = {}

        pools_amounts_out: List[Union[UniswapV2PoolSwapAmounts, UniswapV3PoolSwapAmounts]] = []

        _token_in_quantity: int = 0
        _token_out_quantity: int = 0

        for i, (pool, pool_vector) in enumerate(zip(self.swap_pools, self._swap_vectors)):
            token_in = pool_vector.token_in
            zero_for_one = pool_vector.zero_for_one

            if i == 0:
                _token_in_quantity = token_in_quantity
            else:
                _token_in_quantity = _token_out_quantity

            try:
                if isinstance(pool, LiquidityPool):
                    pool_state_override = pool_state_overrides.get(pool.address)
                    if TYPE_CHECKING:
                        assert pool_state_override is None or isinstance(
                            pool_state_override,
                            UniswapV2PoolState,
                        )
                    _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                        token_in=token_in,
                        token_in_quantity=_token_in_quantity,
                        override_state=pool_state_override,
                    )
                elif isinstance(pool, V3LiquidityPool):
                    pool_state_override = pool_state_overrides.get(pool.address)
                    if TYPE_CHECKING:
                        assert pool_state_override is None or isinstance(
                            pool_state_override,
                            UniswapV3PoolState,
                        )
                    _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                        token_in=token_in,
                        token_in_quantity=_token_in_quantity,
                        override_state=pool_state_override,
                    )
                else:
                    raise ValueError(f"Could not determine Uniswap version for pool {pool}")
            except LiquidityPoolError as e:
                raise ArbitrageError(f"(calculate_tokens_out_from_tokens_in): {e}")
            else:
                if _token_out_quantity == 0:
                    raise ArbitrageError(f"Zero-output swap through pool {pool} @ {pool.address}")

            if isinstance(pool, LiquidityPool):
                pools_amounts_out.append(
                    UniswapV2PoolSwapAmounts(
                        amounts=(0, _token_out_quantity)
                        if zero_for_one
                        else (_token_out_quantity, 0),
                    )
                )
            elif isinstance(pool, V3LiquidityPool):
                pools_amounts_out.append(
                    UniswapV3PoolSwapAmounts(
                        amount_specified=_token_in_quantity,
                        zero_for_one=zero_for_one,
                        sqrt_price_limit_x96=TickMath.MIN_SQRT_RATIO + 1
                        if zero_for_one
                        else TickMath.MAX_SQRT_RATIO - 1,
                    )
                )
            else:
                raise ValueError(
                    f"Could not identify Uniswap version for pool: {self.swap_pools[i]}"
                )

        return pools_amounts_out

    def _update_pool_states(self) -> None:
        """
        Update `self.pool_states` by retrieving `pool.state` from all helpers
        in `self.swap_pools`
        """
        self.pool_states = {pool.address: pool.state for pool in self.swap_pools}

    def auto_update(
        self,
        silent: bool = True,
        block_number: Optional[int] = None,
        override_update_method: Optional[str] = None,
    ) -> bool:
        """
        TBD
        """

        found_updates = False

        if None in self.pool_states.values():
            found_updates = True
            self._update_pool_states()
            self.clear_best()

            return found_updates

        if override_update_method:
            logger.debug(f"OVERRIDDEN UPDATE METHOD: {override_update_method}")

        for pool in self.swap_pools:
            pool_updated = False
            if isinstance(pool, LiquidityPool):
                if pool._update_method == "polling" or override_update_method == "polling":
                    pool_updated = pool.update_reserves(
                        silent=silent,
                        override_update_method=override_update_method,
                        update_block=block_number,
                    )
                elif pool._update_method == "external":
                    if pool.state != self.pool_states[pool.address]:
                        logger.debug(f"(UniswapLpCycle) found update for pool {pool}")
                        pool_updated = True

                if pool_updated:
                    logger.debug(f"(UniswapLpCycle) found update for pool {pool}")
                    found_updates = True
                    break

            elif isinstance(pool, V3LiquidityPool):
                pool_updated, _ = pool.auto_update(
                    silent=silent,
                    block_number=block_number,
                )

                if pool_updated:
                    logger.debug(f"(UniswapLpCycle) found update for pool {pool}")
                    found_updates = True
                    break
            else:
                raise ValueError(f"Could not identify pool {pool}!")

        if found_updates:
            self._update_pool_states()
            self.clear_best()

        return found_updates

    def _pre_calculation_check(
        self,
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ):
        state_overrides = self._sort_overrides(override_state)

        # A scalar value representing the net amount of 1 input token across
        # the complete path (excluding fees).
        # e.g. profit_factor > 1.0 indicates a profitable trade.
        profit_factor: float = 1.0

        # Check the pool state liquidity in the direction of the trade
        for pool, vector in zip(self.swap_pools, self._swap_vectors):
            pool_state = state_overrides.get(pool.address) or pool.state

            if isinstance(pool, LiquidityPool):
                if TYPE_CHECKING:
                    assert isinstance(pool_state, UniswapV2PoolState)

                if pool_state.reserves_token0 == 0 or pool_state.reserves_token1 == 0:
                    raise ZeroLiquidityError(f"V2 pool {pool.address} has no liquidity")

                if pool_state.reserves_token1 == 1 and vector.zero_for_one:
                    raise ZeroLiquidityError(
                        f"V2 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                    )
                elif pool_state.reserves_token0 == 1 and not vector.zero_for_one:
                    raise ZeroLiquidityError(
                        f"V2 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                    )

                price = pool_state.reserves_token1 / pool_state.reserves_token0

            elif isinstance(pool, V3LiquidityPool):
                if TYPE_CHECKING:
                    assert isinstance(pool_state, UniswapV3PoolState)

                if pool_state.sqrt_price_x96 == 0:
                    raise ZeroLiquidityError(
                        f"V3 pool {pool.address} has no liquidity (not initialized)"
                    )

                if pool_state.tick_bitmap == {}:
                    raise ZeroLiquidityError(
                        f"V3 pool {pool.address} has no liquidity (empty bitmap)"
                    )

                if pool_state.liquidity == 0:
                    # Check if the swap is 0 -> 1 and cannot swap any more
                    # token0 for token1
                    if (
                        pool_state.sqrt_price_x96 == TickMath.MIN_SQRT_RATIO + 1
                        and vector.zero_for_one
                    ):
                        raise ZeroLiquidityError(
                            f"V3 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                        )
                    # Check if the swap is 1 -> 0 (zeroForOne=False) and
                    # cannot swap any more token1 for token0
                    elif (
                        pool_state.sqrt_price_x96 == TickMath.MAX_SQRT_RATIO - 1
                        and not vector.zero_for_one
                    ):
                        raise ZeroLiquidityError(
                            f"V3 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                        )

                price = pool_state.sqrt_price_x96**2 / (2**192)

            if isinstance(pool, LiquidityPool):
                # V2 fee is 0.3% by default, represented by 3/1000 = Fraction(3,1000)
                fee = pool.fee_token0 if vector.zero_for_one else pool.fee_token1
            else:
                # V3 fees are integer values representing hundredths of a bip (0.0001)
                # e.g. fee=3000 represents 0.3%
                fee = Fraction(pool._fee, 1000000)

            profit_factor *= (price if vector.zero_for_one else 1 / price) * (
                (fee.denominator - fee.numerator) / fee.denominator
            )

        if profit_factor < 1.0:
            raise ArbitrageError(
                f"No profitable arbitrage at current prices. Profit factor: {profit_factor}"
            )

    def _calculate(
        self,
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ) -> ArbitrageCalculationResult:
        self._pre_calculation_check(override_state)

        state_overrides = self._sort_overrides(override_state)

        # bound the amount to be swapped
        bounds: Tuple[float, float] = (
            1.0,
            float(self.max_input),
        )

        # bracket the initial guess for the algo
        bracket_amount: int = self.max_input
        bracket: Tuple[float, float, float] = (
            0.45 * bracket_amount,
            0.50 * bracket_amount,
            0.55 * bracket_amount,
        )

        def arb_profit(x) -> float:
            token_in_quantity = int(x)  # round the input down
            token_out_quantity: int = 0

            for i, pool in enumerate(self.swap_pools):
                pool_override = state_overrides.get(pool.address)

                if TYPE_CHECKING:
                    assert isinstance(pool, LiquidityPool) and (
                        pool_override is None or isinstance(pool_override, UniswapV2PoolState)
                    )
                    assert isinstance(pool, V3LiquidityPool) and (
                        pool_override is None or isinstance(pool_override, UniswapV3PoolState)
                    )

                try:
                    token_in = self._swap_vectors[i].token_in
                    token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                        token_in=token_in,
                        token_in_quantity=token_in_quantity if i == 0 else token_out_quantity,
                        override_state=pool_override,
                    )
                except (EVMRevertError, LiquidityPoolError):
                    # The optimizer might send invalid amounts into the swap
                    # calculation during iteration. We don't want it to stop,
                    # so catch the exception and pretend the swap results in
                    # token_out_quantity = 0.
                    token_out_quantity = 0
                    break

            # minimize_scalar requires the function to have a minimum value
            # for the solver to settle on an optimum input, so return the
            # negated profit
            return -float(token_out_quantity - token_in_quantity)

        opt = minimize_scalar(
            fun=arb_profit,
            method="bounded",
            bounds=bounds,
            bracket=bracket,
            options={"xatol": 1.0},
        )

        # Negate the result to convert to a sensible value (positive profit)
        best_profit = -int(opt.fun)
        swap_amount = int(opt.x)

        try:
            best_amounts = self._build_amounts_out(
                token_in=self.input_token,
                token_in_quantity=swap_amount,
                pool_state_overrides=state_overrides,
            )
        # except (EVMRevertError, LiquidityPoolError) as e:
        except ArbitrageError as e:
            # Simulated EVM reverts inside the ported `swap` function were
            # ignored to execute the optimizer to completion. Now the optimal
            # value should be tested and raise an exception if it would
            # generate a bad payload that will revert
            raise ArbitrageError(f"No possible arbitrage: {e}") from None
        except Exception as e:
            raise ArbitrageError(f"No possible arbitrage: {e}") from e

        return ArbitrageCalculationResult(
            id=self.id,
            input_token=self.input_token,
            profit_token=self.input_token,
            input_amount=swap_amount,
            profit_amount=best_profit,
            swap_amounts=best_amounts,
        )

    def calculate(
        self,
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ) -> ArbitrageCalculationResult:
        """
        Stateless calculation that does not use `self.best`
        """

        self._pre_calculation_check(override_state)

        return self._calculate(override_state=override_state)

    async def calculate_with_pool(
        self,
        executor: "Executor",
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ) -> asyncio.Future:
        """
        Wrap the arbitrage calculation into an asyncio future using the
        specified executor.

        Arguments
        ---------
        executor : Executor
            An executor (from `concurrent.futures`) to process the calculation
            work. Both `ThreadPoolExecutor` and `ProcessPoolExecutor` are
            supported, but `ProcessPoolExecutor` is recommended.
        override_state : StateOverrideTypes, optional
            An sequence of tuples, representing an ordered pair of helper
            objects for Uniswap V2 / V3 pools and their overridden states.

        Returns
        -------
        A future which returns a `ArbitrageCalculationResult` (or exception)
        when awaited.

        Notes
        -----
        This is an async function that must be called with the `await` keyword.
        """

        self._pre_calculation_check(override_state)

        for pool in self.swap_pools:
            if isinstance(pool, V3LiquidityPool) and pool._sparse_bitmap:
                raise ValueError(
                    f"Cannot process arbitrage {self} with executor: {pool.address} has sparse bitmap"
                )

        return asyncio.get_running_loop().run_in_executor(
            executor,
            self.calculate,
            override_state,
        )

    def calculate_arbitrage_return_best(
        self,
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ):
        """
        A wrapper over `calculate_arbitrage`, useful for sending the
        calculation into a process pool and retrieving the results after
        pickling/unpickling the object and losing connection to the original.
        """

        self.calculate_arbitrage(override_state)
        return self.id, self.best

    def calculate_arbitrage(
        self,
        override_state: Optional[
            Sequence[
                Union[
                    Tuple[LiquidityPool, UniswapV2PoolState],
                    Tuple[LiquidityPool, UniswapV2PoolSimulationResult],
                    Tuple[V3LiquidityPool, UniswapV3PoolState],
                    Tuple[V3LiquidityPool, UniswapV3PoolSimulationResult],
                ]
            ]
        ] = None,
    ) -> Tuple[bool, Tuple[int, int]]:
        """
        TBD
        """

        result = self._calculate(override_state=override_state)

        if override_state is None:
            self.best.update(
                {
                    "last_swap_amount": result.input_amount,
                    "profit_amount": result.profit_amount,
                    "swap_amount": result.input_amount,
                    "swap_pool_amounts": result.swap_amounts,
                }
            )

        profitable = result.profit_amount > 0
        return profitable, (result.input_amount, result.profit_amount)

    def clear_best(self):
        self.best.update(
            {
                "profit_amount": 0,
                "swap_amount": 0,
                "swap_pool_amounts": [],
            }
        )

    @classmethod
    def from_addresses(
        cls,
        input_token_address: str,
        swap_pool_addresses: Iterable[Tuple[str, str]],
        id: str,
        max_input: Optional[int] = None,
    ) -> "UniswapLpCycle":
        """
        Create a new `UniswapLpCycle` object from token and pool addresses.

        Arguments
        ---------
        input_token_address : str
            A address for the input_token
        swap_pool_addresses : Iterable[str,str]
            An iterable of tuples representing the address for each pool in the
            swap path, and a string specifying the Uniswap version for that
            pool (either "V2" or "V3")

            e.g. swap_pool_addresses = [
                ("0xCBCdF9626bC03E24f779434178A73a0B4bad62eD","V3"),
                ("0xbb2b8038a1640196fbe3e38816f3e67cba72d940","V2")
            ]
        max_input: int, optional
            The maximum input amount for the input token (limited by the
            balance of the deployed contract or operating EOA)
        id: str, optional
            A unique identifier for bookkeeping purposes, not validated
        """

        # create the token object
        token = Erc20Token(input_token_address)

        # create the pool objects
        pool_objects: List[
            Union[
                LiquidityPool,
                V3LiquidityPool,
                CamelotLiquidityPool,
            ]
        ] = []
        for pool_address, pool_type in swap_pool_addresses:
            if pool_type == "V2":
                pool_objects.append(LiquidityPool(address=pool_address))
            elif pool_type == "V3":
                pool_objects.append(V3LiquidityPool(address=pool_address))
            elif pool_type == "CamelotV2":
                pool_objects.append(CamelotLiquidityPool(address=pool_address))
            else:
                raise ArbitrageError(f"Pool type {pool_type} unknown!")

        return cls(
            input_token=token,
            swap_pools=pool_objects,
            max_input=max_input,
            id=id,
        )

    def generate_payloads(
        self,
        from_address: Union[str, ChecksumAddress],
        swap_amount: Optional[int] = None,
        pool_swap_amounts: Optional[
            Sequence[
                Union[
                    UniswapV2PoolSwapAmounts,
                    UniswapV3PoolSwapAmounts,
                ]
            ]
        ] = None,
    ) -> List[Tuple[ChecksumAddress, bytes, int]]:
        """
        Generate a list of ABI-encoded calldata for each step in the swap path.

        Calldata is built using the eth_abi.encode method and the ABI for the
        ``swap`` function at the Uniswap pool. V2 and V3 pools are supported.

        Arguments
        ---------
        from_address: str
            The address that will execute the calldata. Must be a smart
            contract implementing the required callbacks specific to the pool.

        swap_amount: int, optional
            The initial amount of `token_in` to swap through the first pool.
            If this argument is `None`, amount will be retrieved from
            `self.best`.

        pool_swap_amounts: Iterable[UniswapV2PoolSwapAmounts |
        UniswapV3PoolSwapAmounts], optional
            An iterable of swap amounts to be encoded. If this argument is
            `None`, amounts will be retrieved from `self.best`.

        Returns
        -------
        ``list[(str, bytes, int)]``
            A list of payload tuples. Each payload tuple has form (
            address: ChecksumAddress, calldata: bytes, value: int).

        Raises
        ------
        ArbitrageError
            if the generated payloads would revert on-chain, or if the inputs
            were invalid
        """

        from_address = to_checksum_address(from_address)

        if swap_amount is None:
            swap_amount = self.best["swap_amount"]

        if pool_swap_amounts is None:
            pool_swap_amounts = self.best["swap_pool_amounts"]

        # Abandon empty inputs.
        # @dev this looks like a useful place for a ValueError, but threaded
        # clients may execute a pool update for a swap pool before the call to
        # generate payloads is processed. Abandon the call in this case and
        # raise a generic non-fatal exception.
        if not pool_swap_amounts:
            raise ArbitrageError("Pool amounts empty, abandoning payload generation.")

        payloads = []
        msg_value: int = (
            0  # This arbitrage does not require a `msg.value` payment
        )

        first_pool = self.swap_pools[0]
        last_pool = self.swap_pools[-1]

        try:
            if isinstance(first_pool, LiquidityPool):
                # Special case: If first pool is type V2, input token must be
                # transferred prior to the swap
                payloads.append(
                    (
                        # address
                        self.input_token.address,
                        # bytes calldata
                        Web3.keccak(text="transfer(address,uint256)")[:4]
                        + eth_abi.encode(
                            types=(
                                "address",
                                "uint256",
                            ),
                            args=(
                                first_pool.address,
                                swap_amount,
                            ),
                        ),
                        msg_value,
                    )
                )

            for i, (swap_pool, _swap_amounts) in enumerate(zip(self.swap_pools, pool_swap_amounts)):
                if swap_pool is last_pool:
                    next_pool = None
                else:
                    next_pool = self.swap_pools[i + 1]

                if next_pool is not None:
                    # V2 pools require a pre-swap transfer, so the contract
                    # does not have to perform intermediate custody and the
                    # swap can send the tokens directly to the next pool
                    if isinstance(next_pool, LiquidityPool):
                        swap_destination_address = next_pool.address
                    # V3 pools cannot accept a pre-swap transfer, so the contract
                    # must maintain custody prior to a swap
                    elif isinstance(next_pool, V3LiquidityPool):
                        swap_destination_address = from_address
                else:
                    # Set the destination address for the last swap to the
                    # sending address
                    swap_destination_address = from_address

                if isinstance(swap_pool, LiquidityPool):
                    if TYPE_CHECKING:
                        assert isinstance(_swap_amounts, UniswapV2PoolSwapAmounts)
                    logger.debug(f"PAYLOAD: building V2 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                            + eth_abi.encode(
                                types=(
                                    "uint256",
                                    "uint256",
                                    "address",
                                    "bytes",
                                ),
                                args=(
                                    *_swap_amounts.amounts,
                                    swap_destination_address,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                elif isinstance(swap_pool, V3LiquidityPool):
                    if TYPE_CHECKING:
                        assert isinstance(_swap_amounts, UniswapV3PoolSwapAmounts)
                    logger.debug(f"PAYLOAD: building V3 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(address,bool,int256,uint160,bytes)")[:4]
                            + eth_abi.encode(
                                types=(
                                    "address",
                                    "bool",
                                    "int256",
                                    "uint160",
                                    "bytes",
                                ),
                                args=(
                                    swap_destination_address,
                                    _swap_amounts.zero_for_one,
                                    _swap_amounts.amount_specified,
                                    _swap_amounts.sqrt_price_limit_x96,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                else:
                    raise ValueError(
                        f"Could not identify pool: {swap_pool}, type={type(swap_pool)}"
                    )
        except Exception as e:
            logger.exception("generate_payloads catch-all")
            raise ArbitrageError(f"generate_payloads (catch-all)): {e}") from e

        return payloads



from typing import Any, Dict, Optional, Tuple
from warnings import warn

import ujson  # type: ignore[import]
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3.exceptions import BadFunctionCallOutput, ContractLogicError

from .chainlink import ChainlinkPriceContract
from .config import get_web3
from .logging import logger

# Taken from OpenZeppelin's ERC-20 implementation
# ref: https://www.npmjs.com/package/@openzeppelin/contracts?activeTab=code
ERC20_ABI_MINIMAL = ujson.loads(
    '[{"inputs": [{"internalType": "string", "name": "name_", "type": "string"}, {"internalType": "string", "name": "symbol_", "type": "string"}], "stateMutability": "nonpayable", "type": "constructor"}, {"anonymous": false, "inputs": [{"indexed": true, "internalType": "address", "name": "owner", "type": "address"}, {"indexed": true, "internalType": "address", "name": "spender", "type": "address"}, {"indexed": false, "internalType": "uint256", "name": "value", "type": "uint256"}], "name": "Approval", "type": "event"}, {"anonymous": false, "inputs": [{"indexed": true, "internalType": "address", "name": "from", "type": "address"}, {"indexed": true, "internalType": "address", "name": "to", "type": "address"}, {"indexed": false, "internalType": "uint256", "name": "value", "type": "uint256"}], "name": "Transfer", "type": "event"}, {"inputs": [{"internalType": "address", "name": "owner", "type": "address"}, {"internalType": "address", "name": "spender", "type": "address"}], "name": "allowance", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "approve", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "account", "type": "address"}], "name": "balanceOf", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "decimals", "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "subtractedValue", "type": "uint256"}], "name": "decreaseAllowance", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "addedValue", "type": "uint256"}], "name": "increaseAllowance", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [], "name": "name", "outputs": [{"internalType": "string", "name": "", "type": "string"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "symbol", "outputs": [{"internalType": "string", "name": "", "type": "string"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "totalSupply", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "to", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "transfer", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "from", "type": "address"}, {"internalType": "address", "name": "to", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "transferFrom", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}]'
)


class Erc20Token:
    """
    An ERC-20 token contract.

    If an ABI is specified, it will be used. Otherwise, a minimal ERC-20 ABI
    will be used that provides access to the interface defined by EIP-20, plus
    some commonly used extensions:

    - Functions:
        # EIP-20 BASELINE
        - totalSupply()
        - balanceOf(account)
        - transfer(to, amount)
        - allowance(owner, spender)
        - approve(spender, amount)
        - transferFrom(from, to, amount)
        # METADATA EXTENSIONS
        - name()
        - symbol()
        - decimals()
        # ALLOWANCE EXTENSIONS
        - increaseAllowance(spender, addedValue)
        - decreaseAllowance(spender, subtractedValue)
    - Events:
        - Transfer(from, to, value)
        - Approval(owner, spender, value)
    """

    __slots__: Tuple[str, ...] = (
        "address",
        "abi",
        "_price_oracle",
        "_w3",
        "_w3_contract",
        "name",
        "symbol",
        "decimals",
        "price",
    )

    def __init__(
        self,
        address: str,
        abi: Optional[list] = None,
        oracle_address: Optional[str] = None,
        silent: bool = False,
        unload_brownie_contract_after_init: bool = False,  # deprecated
        min_abi: bool = False,  # deprecated
        user: Optional[Any] = None,  # deprecated
    ) -> None:
        self.address: ChecksumAddress = to_checksum_address(address)
        self.abi = abi or ERC20_ABI_MINIMAL

        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        self._w3_contract = self._w3.eth.contract(
            address=self.address,
            abi=self.abi,
        )

        if user:
            warn(
                "Instantiating with a single user is deprecated. You may use "
                "the get_balance() method to retrieve token balances for a "
                "particular address."
            )
            # self._user = user

        if min_abi:
            warn(
                "Using a minimal ABI is now the default behavior. Remove "
                "min_abi constructor argument to stop seeing this message."
            )

        if unload_brownie_contract_after_init:
            warn(
                "unload_brownie_contract_after_init is no longer needed and is "
                "ignored. Remove constructor argument to stop seeing this "
                "message."
            )

        try:
            self.name: str
            self.name = self._w3_contract.functions.name().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            # Workaround for non-ERC20 compliant tokens
            for func in ("name", "NAME"):
                try:
                    self.name = (
                        self._w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        )
                    ).decode("utf-8", errors="ignore")
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.name @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.name
        except AttributeError:
            if not self._w3.eth.get_code(self.address):
                raise ValueError("No contract deployed at this address")
            self.name = f"Unknown @ {self.address}"
            warn(
                f"Token contract at {address} does not implement a 'name' function."
            )

        try:
            self.symbol: str
            self.symbol = self._w3_contract.functions.symbol().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            for func in ("symbol", "SYMBOL"):
                # Workaround for non-ERC20 compliant tokens
                try:
                    self.symbol = (
                        self._w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        )
                    ).decode("utf-8", errors="ignore")
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.symbol @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.symbol
        except AttributeError:
            if not self._w3.eth.get_code(self.address):
                raise ValueError("No contract deployed at this address")
            self.symbol = "UNKNOWN"
            warn(
                f"Token contract at {address} does not implement a 'symbol' function."
            )

        try:
            self.decimals: int
            self.decimals = self._w3_contract.functions.decimals().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            for func in ("decimals", "DECIMALS"):
                try:
                    # Workaround for non-ERC20 compliant tokens
                    self.decimals = int(
                        self._w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        ).hex(),
                        16,
                    )
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.decimals @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.decimals
        except Exception:
            if not self._w3.eth.get_code(self.address):
                raise ValueError("No contract deployed at this address")
            self.decimals = 0
            warn(
                f"Token contract at {address} does not implement a 'decimals' function. Setting to 0."
            )

        # if user:
        #     self.update_balance(user)
        #     # self.balance = self._brownie_contract.balanceOf(self._user)
        #     self.balance = self.get_balance(self.address)
        #     self.normalized_balance = self.balance / (10**self.decimals)

        self.price: Optional[float]
        if oracle_address:
            self._price_oracle = ChainlinkPriceContract(address=oracle_address)
            self.price = self._price_oracle.price
        else:
            self.price = None

        if not silent:
            logger.info(f"• {self.symbol} ({self.name})")

    # Web3 objects and Brownie contracts cannot be pickled
    def __getstate__(self):
        keys_to_remove = (
            "_w3",
            "_w3_contract",
        )

        try:
            self.__slots__
        except AttributeError:
            pass
        else:
            return {
                attr_name: getattr(self, attr_name, None)
                for attr_name in self.__slots__
                if attr_name not in keys_to_remove
            }

        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in keys_to_remove
        }

    def __setstate__(self, state: Dict):
        for key, value in state.items():
            setattr(self, key, value)

    def __eq__(self, other) -> bool:
        if isinstance(other, Erc20Token):
            return self.address == other.address
        elif isinstance(other, str):
            return self.address.lower() == other.lower()
        else:
            return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, Erc20Token):
            return self.address < other.address
        elif isinstance(other, str):
            return self.address.lower() < other.lower()
        else:
            return NotImplemented

    def __gt__(self, other) -> bool:
        if isinstance(other, Erc20Token):
            return self.address > other.address
        elif isinstance(other, str):
            return self.address.lower() > other.lower()
        else:
            return NotImplemented

    def __str__(self):
        return self.symbol

    def get_approval(self, owner: str, spender: str) -> int:
        return self._w3_contract.functions.allowance(
            to_checksum_address(owner),
            to_checksum_address(spender),
        ).call()

    # def set_approval(self, owner: str, spender: str, value: Union[int, str]):
    #     """
    #     Sets the approval value for an external contract to transfer tokens
    #     quantites up to the specified amount from this address. For unlimited
    #     approval, set value to the string "UNLIMITED".
    #     """

    #     if isinstance(value, int):
    #         if not (MIN_UINT256 <= value <= MAX_UINT256):
    #             raise ValueError(
    #                 f"Provide an integer value between 0 and 2**256-1"
    #             )
    #     elif isinstance(value, str):
    #         if value != "UNLIMITED":
    #             raise ValueError("Value must be 'UNLIMITED' or an integer")
    #         else:
    #             print("Setting unlimited approval!")
    #             value = MAX_UINT256
    #     else:
    #         raise TypeError(
    #             f"Value must be an integer or string! Was {type(value)}"
    #         )

    #     try:
    #         self._w3_contract.functions["approve"](
    #             to_checksum_address(owner),
    #             to_checksum_address(spender),
    #             value,
    #         )
    #         # self._brownie_contract.approve(
    #         #     external_address,
    #         #     value,
    #         #     {"from": self._user.address},
    #         # )
    #     except Exception as e:
    #         print(f"Exception in token_approve: {e}")
    #         raise

    def get_balance(self, address: str) -> int:
        return self._w3_contract.functions.balanceOf(
            to_checksum_address(address)
        ).call()

    # def update_balance(self):
    #     self.balance = self.get_balance(self._user)
    #     self.normalized_balance = self.balance / (10**self.decimals)

    def update_price(self):
        self._price_oracle.update_price()
        self.price = self._price_oracle.price



from io import TextIOWrapper
from typing import Dict, List, Optional, TextIO, Tuple, Union

import ujson  # type: ignore[import]
from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3._utils.events import get_event_data
from web3._utils.filters import construct_event_filter_params

from ..config import get_web3
from ..logging import logger
from .abi import UNISWAP_V3_POOL_ABI
from .v3_dataclasses import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolExternalUpdate,
)
from .v3_liquidity_pool import V3LiquidityPool


class UniswapV3LiquiditySnapshot:
    """
    Retrieve and maintain liquidity positions for Uniswap V3 pools.
    """

    def __init__(
        self,
        file: Union[TextIO, str],
        chain_id: Optional[int] = None,
    ):
        _file: TextIOWrapper
        json_liquidity_snapshot: dict

        try:
            if isinstance(file, TextIOWrapper):
                _file = file
                json_liquidity_snapshot = ujson.load(file)
            elif isinstance(file, str):
                with open(file) as _file:
                    json_liquidity_snapshot = ujson.load(_file)
            else:
                raise ValueError(f"GOT {type(file)}")
        except:
            raise
        finally:
            _file.close()

        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        self._chain_id = chain_id or self._w3.eth.chain_id

        self.newest_block = json_liquidity_snapshot.pop("snapshot_block")

        self._liquidity_snapshot: Dict[ChecksumAddress, Dict] = dict()
        for (
            pool_address,
            pool_liquidity_snapshot,
        ) in json_liquidity_snapshot.items():
            self._liquidity_snapshot[to_checksum_address(pool_address)] = {
                "tick_bitmap": {
                    int(k): UniswapV3BitmapAtWord(**v)
                    for k, v in pool_liquidity_snapshot["tick_bitmap"].items()
                },
                "tick_data": {
                    int(k): UniswapV3LiquidityAtTick(**v)
                    for k, v in pool_liquidity_snapshot["tick_data"].items()
                },
            }

        logger.info(
            f"Loaded LP snapshot: {len(json_liquidity_snapshot)} pools @ block {self.newest_block}"
        )

        self._liquidity_events: Dict[
            ChecksumAddress, List[UniswapV3LiquidityEvent]
        ] = dict()

    def _add_pool_if_missing(self, pool_address: ChecksumAddress) -> None:
        try:
            self._liquidity_events[pool_address]
        except KeyError:
            self._liquidity_events[pool_address] = []

        try:
            self._liquidity_snapshot[pool_address]
        except KeyError:
            self._liquidity_snapshot[pool_address] = {}

    def fetch_new_liquidity_events(
        self,
        to_block: int,
        span: int = 1000,
    ) -> None:
        def _process_log() -> Tuple[ChecksumAddress, UniswapV3LiquidityEvent]:
            decoded_event = get_event_data(self._w3.codec, event_abi, log)

            pool_address = to_checksum_address(decoded_event["address"])
            tx_index = decoded_event["transactionIndex"]
            liquidity_block = decoded_event["blockNumber"]
            liquidity = decoded_event["args"]["amount"] * (
                -1 if decoded_event["event"] == "Burn" else 1
            )
            tick_lower = decoded_event["args"]["tickLower"]
            tick_upper = decoded_event["args"]["tickUpper"]

            return pool_address, UniswapV3LiquidityEvent(
                block_number=liquidity_block,
                liquidity=liquidity,
                tick_lower=tick_lower,
                tick_upper=tick_upper,
                tx_index=tx_index,
            )

        logger.info(
            f"Updating snapshot from block {self.newest_block} to {to_block}"
        )

        v3pool = Web3().eth.contract(abi=UNISWAP_V3_POOL_ABI)

        for event in [v3pool.events.Mint, v3pool.events.Burn]:
            logger.info(f"Processing {event.event_name} events")
            event_abi = event._get_event_abi()
            start_block = self.newest_block + 1

            while True:
                end_block = min(to_block, start_block + span - 1)

                _, event_filter_params = construct_event_filter_params(
                    event_abi=event_abi,
                    abi_codec=self._w3.codec,
                    fromBlock=start_block,
                    toBlock=end_block,
                )

                event_logs = self._w3.eth.get_logs(event_filter_params)

                for log in event_logs:
                    pool_address, liquidity_event = _process_log()

                    # skip zero liquidity events
                    if liquidity_event.liquidity == 0:
                        continue

                    self._add_pool_if_missing(pool_address)
                    self._liquidity_events[pool_address].append(
                        liquidity_event
                    )

                if end_block == to_block:
                    break
                else:
                    start_block = end_block + 1

        logger.info(f"Updated snapshot to block {to_block}")
        self.newest_block = to_block

    def get_pool_updates(
        self, pool_address
    ) -> List[UniswapV3PoolExternalUpdate]:
        try:
            self._liquidity_events[pool_address]
        except KeyError:
            return []
        else:
            # Sort the liquidity events by block, then transaction index
            # before returning them.
            # @dev the V3LiquidityPool helper will reject liquidity events
            # associated with a past block, so they must be processed in
            # chronological order
            sorted_events = sorted(
                self._liquidity_events[pool_address],
                key=lambda event: (event.block_number, event.tx_index),
            )
            self._liquidity_events[pool_address].clear()

            return [
                UniswapV3PoolExternalUpdate(
                    block_number=event.block_number,
                    liquidity_change=(
                        event.liquidity,
                        event.tick_lower,
                        event.tick_upper,
                    ),
                )
                for event in sorted_events
            ]

    def get_tick_bitmap(
        self, pool: Union[ChecksumAddress, V3LiquidityPool]
    ) -> Dict[int, UniswapV3BitmapAtWord]:
        if isinstance(pool, V3LiquidityPool):
            pool_address = pool.address
        elif isinstance(pool, str):
            pool_address = to_checksum_address(pool)
        else:
            raise ValueError(f"Unexpected input for pool: {type(pool)}")

        try:
            return self._liquidity_snapshot[pool_address]["tick_bitmap"]
        except KeyError:
            return {}

    def get_tick_data(
        self, pool: Union[ChecksumAddress, V3LiquidityPool]
    ) -> Dict[int, UniswapV3LiquidityAtTick]:
        if isinstance(pool, V3LiquidityPool):
            pool_address = pool.address
        elif isinstance(pool, str):
            pool_address = to_checksum_address(pool)
        else:
            raise ValueError(f"Unexpected input for pool: {type(pool)}")

        try:
            return self._liquidity_snapshot[pool_address]["tick_data"]
        except KeyError:
            return {}

    def update_snapshot(
        self,
        pool: Union[V3LiquidityPool, ChecksumAddress],
        tick_data: Dict[int, UniswapV3LiquidityAtTick],
        tick_bitmap: Dict[int, UniswapV3BitmapAtWord],
    ) -> None:
        if isinstance(pool, V3LiquidityPool):
            pool_address = pool.address
        elif isinstance(pool, str):
            pool_address = to_checksum_address(pool)
        else:
            raise ValueError(f"Unexpected input for pool: {type(pool)}")

        self._add_pool_if_missing(pool_address)
        self._liquidity_snapshot[pool_address].update(
            {
                "tick_bitmap": tick_bitmap,
            }
        )
        self._liquidity_snapshot[pool_address].update(
            {
                "tick_data": tick_data,
            }
        )



from threading import Lock
from typing import TYPE_CHECKING, Dict, Optional

from eth_utils import to_checksum_address

from ..config import get_web3
from ..exceptions import ManagerError
from ..erc20_token import Erc20Token
from ..baseclasses import HelperManager

if TYPE_CHECKING:
    from ..baseclasses import TokenHelper

_all_tokens: Dict[
    int,
    Dict[str, "TokenHelper"],
] = {}


class AllTokens:
    def __init__(self, chain_id):
        try:
            _all_tokens[chain_id]
        except KeyError:
            _all_tokens[chain_id] = {}
        finally:
            self.tokens = _all_tokens[chain_id]

    def __delitem__(self, token_address: str):
        del self.tokens[token_address]

    def __getitem__(self, token_address: str):
        return self.tokens[token_address]

    def __setitem__(
        self,
        token_address: str,
        token_helper: "TokenHelper",
    ):
        self.tokens[token_address] = token_helper

    def __len__(self):
        return len(self.tokens)

    def get(self, token_address: str):
        try:
            return self.tokens[token_address]
        except KeyError:
            return None


class Erc20TokenHelperManager(HelperManager):
    """
    A class that generates and tracks Erc20Token helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    _state: dict = {}

    def __init__(self, chain_id: Optional[int] = None):
        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        chain_id = chain_id or self._w3.eth.chain_id

        # the internal state data for this object is held in the
        # class-level _state dictionary, keyed by the chain ID
        if self._state.get(chain_id):
            self.__dict__ = self._state[chain_id]
        else:
            self._state[chain_id] = {}
            self.__dict__ = self._state[chain_id]

            # initialize internal attributes
            self._erc20tokens: dict = {}
            self._lock = Lock()

    def get_erc20token(
        self,
        address: str,
        # accept any number of keyword arguments, which are
        # passed directly to Erc20Token without validation
        **kwargs,
    ) -> Erc20Token:
        """
        Get the token object from its address
        """

        address = to_checksum_address(address)

        if token_helper := self._erc20tokens.get(address):
            return token_helper

        try:
            token_helper = Erc20Token(address=address, **kwargs)
        except Exception:
            raise ManagerError(
                f"Could not create Erc20Token helper: {address=}"
            )

        with self._lock:
            self._erc20tokens[address] = token_helper

        return token_helper



from threading import Lock
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple, Union

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address
from web3 import Web3

from ..config import get_web3
from ..constants import ZERO_ADDRESS
from ..dex.uniswap import FACTORY_ADDRESSES, TICKLENS_ADDRESSES
from ..exceptions import ManagerError, PoolNotAssociated
from ..logging import logger
from ..manager import AllPools, Erc20TokenHelperManager
from ..erc20_token import Erc20Token
from ..baseclasses import HelperManager
from .abi import UNISWAP_V2_FACTORY_ABI
from .v2_liquidity_pool import LiquidityPool
from .v3_functions import generate_v3_pool_address
from .v3_snapshot import UniswapV3LiquiditySnapshot
from .v3_tick_lens import TickLens
from .v3_liquidity_pool import V3LiquidityPool


class UniswapLiquidityPoolManager(HelperManager):
    """
    Single-concern base class to allow derived classes to share state
    """

    _state: Dict[int, Dict] = dict()

    def __init__(
        self,
        factory_address: str,
        chain_id: int,
    ):
        """
        Initialize the specific state dictionary for the given chain id and
        factory address
        """

        # the internal state data for all child objects is held in a nested
        # class-level dictionary, keyed by chain ID and factory address
        try:
            self._state[chain_id]
        except KeyError:
            self._state[chain_id] = {}
            self._state[chain_id][
                "erc20token_manager"
            ] = Erc20TokenHelperManager(chain_id)

        try:
            self._state[chain_id][factory_address]
        except KeyError:
            self._state[chain_id][factory_address] = {}

    @classmethod
    def add_chain(cls, chain_id: int) -> None:
        """
        Add a new chain ID.
        """
        if not FACTORY_ADDRESSES.get(chain_id):
            FACTORY_ADDRESSES[chain_id] = {}

    @classmethod
    def add_factory(cls, chain_id: int, factory_address: str) -> None:
        """
        Add a new factory address at a given chain ID.
        """
        cls.add_chain(chain_id=chain_id)

        factory_address = to_checksum_address(factory_address)

        if not FACTORY_ADDRESSES[chain_id].get(factory_address):
            FACTORY_ADDRESSES[chain_id][factory_address] = {}

    @classmethod
    def add_pool_init_hash(
        cls, chain_id: int, factory_address: str, pool_init_hash: str
    ):
        """
        Add a pool_init_hash for a factory at a given chain ID.
        """

        factory_address = to_checksum_address(factory_address)

        cls.add_factory(chain_id=chain_id, factory_address=factory_address)

        if not FACTORY_ADDRESSES[chain_id][factory_address].get("init_hash"):
            FACTORY_ADDRESSES[chain_id][factory_address][
                "init_hash"
            ] = pool_init_hash


class UniswapV2LiquidityPoolManager(UniswapLiquidityPoolManager):
    """
    A class that generates and tracks Uniswap V2 liquidity pool helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    def __init__(
        self,
        factory_address: Union[ChecksumAddress, str],
        chain_id: Optional[int] = None,
    ):
        _web3 = get_web3()
        if _web3 is not None:
            pass
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                _web3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        if TYPE_CHECKING:
            assert isinstance(_web3, Web3)

        chain_id = chain_id or _web3.eth.chain_id

        factory_address = to_checksum_address(factory_address)

        super().__init__(
            factory_address=factory_address,
            chain_id=chain_id,
        )

        self.__dict__ = self._state[chain_id][factory_address]

        if self.__dict__ == {}:
            self._w3 = _web3
            self.chain_id = chain_id
            self._factory_address = factory_address
            self._w3_contract = self._w3.eth.contract(
                address=factory_address, abi=UNISWAP_V2_FACTORY_ABI
            )
            self._lock = Lock()
            self._tracked_pools: Dict[ChecksumAddress, LiquidityPool] = dict()
            self._token_manager: Erc20TokenHelperManager = self._state[
                chain_id
            ]["erc20token_manager"]
            self._factory_init_hash = FACTORY_ADDRESSES[chain_id][
                self._factory_address
            ]["init_hash"]
            self._untracked_pools: Set[ChecksumAddress] = set()

    def __delitem__(
        self, pool: Union[LiquidityPool, str, ChecksumAddress]
    ) -> None:
        pool_address: ChecksumAddress

        if isinstance(pool, LiquidityPool):
            pool_address = pool.address
        else:
            pool_address = to_checksum_address(pool)

        try:
            del self._tracked_pools[pool_address]
        except KeyError:
            pass

    def __repr__(self):
        return (
            f"UniswapV2LiquidityPoolManager(factory={self._factory_address})"
        )

    def _add_pool(self, pool_helper: LiquidityPool):
        with self._lock:
            self._tracked_pools[pool_helper.address] = pool_helper

    def get_pool(
        self,
        pool_address: Optional[str] = None,
        token_addresses: Optional[Tuple[str, str]] = None,
        silent: bool = False,
        update_method: str = "polling",
        state_block: Optional[int] = None,
    ) -> LiquidityPool:
        """
        Get the pool object from its address, or a tuple of token addresses
        """

        # WIP: re-work to avoid storing pools by token, convert to address first

        pool_helper: LiquidityPool

        if token_addresses is not None:
            if len(token_addresses) != 2:
                raise ValueError("Provide exactly two token addresses")

            checksummed_token_addresses = tuple(
                [
                    to_checksum_address(token_address)
                    for token_address in token_addresses
                ]
            )

            try:
                for token_address in checksummed_token_addresses:
                    self._token_manager.get_erc20token(
                        address=token_address,
                        silent=silent,
                    )
            except Exception:
                raise ManagerError("Could not get both Erc20Token helpers")

            pool_address = to_checksum_address(
                self._w3_contract.functions.getPair(
                    *checksummed_token_addresses
                ).call()
            )

            if pool_address == ZERO_ADDRESS:
                raise ManagerError("No V2 LP available")

        if TYPE_CHECKING:
            assert pool_address is not None
        # Address is now known, check if the pool is already being tracked
        pool_address = to_checksum_address(pool_address)

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(
                f"Pool address {pool_address} not associated with factory {self._factory_address}"
            )

        try:
            pool_helper = self._tracked_pools[pool_address]
        except KeyError:
            pass
        else:
            assert isinstance(
                pool_helper, LiquidityPool
            ), f"{self} Attempted to return non-V2 pool {pool_helper}! {pool_address=}, {token_addresses=}"
            return pool_helper

        # Check if the AllPools collection already has this pool
        # print(f"(get_pool) no pool found, checking AllPools")
        if pool_helper := AllPools(self.chain_id).get(pool_address):
            if pool_helper.factory == self._factory_address:
                print(
                    "(get_pool) found pool in AllPools, factory address match"
                )
                self._add_pool(pool_helper)
                assert pool_helper.address in self._tracked_pools
                assert isinstance(
                    pool_helper, LiquidityPool
                ), f"{self} Attempted to return non-V2 pool {pool_helper}! {pool_address=}, {token_addresses=}"
                return pool_helper
            else:
                self._untracked_pools.add(pool_address)

        try:
            pool_helper = LiquidityPool(
                address=pool_address,
                silent=silent,
                state_block=state_block,
                factory_address=self._factory_address,
                factory_init_hash=self._factory_init_hash,
                update_method=update_method,
            )
        except Exception as e:
            self._untracked_pools.add(pool_address)
            raise ManagerError(f"Could not build V2 pool {pool_address}: {e}")
        else:
            self._add_pool(pool_helper)
            assert isinstance(
                pool_helper, LiquidityPool
            ), f"{self} Attempted to return non-V2 pool {pool_helper}! {pool_address=}, {token_addresses=}"
            return pool_helper


class UniswapV3LiquidityPoolManager(UniswapLiquidityPoolManager):
    """
    A class that generates and tracks Uniswap V3 liquidity pool helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    def __init__(
        self,
        factory_address: Union[ChecksumAddress, str],
        chain_id: Optional[int] = None,
        snapshot: Optional[UniswapV3LiquiditySnapshot] = None,
    ):
        _web3 = get_web3()
        if _web3 is not None:
            pass
        else:
            from brownie import web3 as brownie_web3  # type: ignore[import]

            if brownie_web3.isConnected():
                _web3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        if TYPE_CHECKING:
            assert isinstance(_web3, Web3)

        chain_id = chain_id or _web3.eth.chain_id

        factory_address = to_checksum_address(factory_address)

        super().__init__(
            factory_address=factory_address,
            chain_id=chain_id,
        )

        self.__dict__ = self._state[chain_id][factory_address]

        if self.__dict__ == {}:
            try:
                self._w3 = _web3
                self.chain_id = chain_id
                self._factory_address = to_checksum_address(factory_address)
                self._lens = TickLens(
                    address=TICKLENS_ADDRESSES[self._w3.eth.chain_id][
                        self._factory_address
                    ]
                )
                self._lock = Lock()
                self._tracked_pools: Dict[
                    ChecksumAddress, V3LiquidityPool
                ] = {}
                self._token_manager: Erc20TokenHelperManager = self._state[
                    chain_id
                ]["erc20token_manager"]
                self._factory_init_hash = FACTORY_ADDRESSES[chain_id][
                    self._factory_address
                ]["init_hash"]
                self._snapshot = snapshot
                self._untracked_pools: Set[ChecksumAddress] = set()
            except:
                self._state[chain_id][factory_address] = {}
                raise

    def __delitem__(
        self, pool: Union[V3LiquidityPool, str, ChecksumAddress]
    ) -> None:
        pool_address: ChecksumAddress

        if isinstance(pool, V3LiquidityPool):
            pool_address = pool.address
        else:
            pool_address = to_checksum_address(pool)

        try:
            del self._tracked_pools[pool_address]
        except KeyError:
            pass

    def __repr__(self):
        return (
            f"UniswapV3LiquidityPoolManager(factory={self._factory_address})"
        )

    def _store_pool(self, pool_helper: V3LiquidityPool):
        with self._lock:
            self._tracked_pools[pool_helper.address] = pool_helper

    def get_pool(
        self,
        pool_address: Optional[Union[ChecksumAddress, str]] = None,
        token_addresses: Optional[
            Tuple[Union[str, ChecksumAddress], Union[str, ChecksumAddress]]
        ] = None,
        pool_fee: Optional[int] = None,
        silent: bool = False,
        # keyword arguments passed to the `V3LiquidityPool` constructor
        v3liquiditypool_kwargs: Optional[dict] = None,
        state_block: Optional[int] = None,
    ) -> V3LiquidityPool:
        """
        Get the pool object from its address, or a tuple of token addresses
        and fee in bips (e.g. 100, 500, 3000, 10000)
        """

        def apply_liquidity_updates(pool: V3LiquidityPool):
            logger.debug(f"Applying liquidity updates to {pool}")
            if not self._snapshot:
                return
            for external_update in self._snapshot.get_pool_updates(
                pool.address
            ):
                pool.external_update(update=external_update, force=True)

        def find_or_build(
            pool_address: ChecksumAddress,
            state_block: Optional[int] = None,
        ) -> V3LiquidityPool:
            if TYPE_CHECKING:
                assert isinstance(v3liquiditypool_kwargs, dict)

            # Check if the AllPools collection already has this pool
            pool_helper: V3LiquidityPool = AllPools(self.chain_id).get(
                pool_address
            )
            if pool_helper:
                if pool_helper.factory == self._factory_address:
                    self._store_pool(pool_helper)
                    assert pool_helper.address in self._tracked_pools
                    assert isinstance(
                        pool_helper, V3LiquidityPool
                    ), f"{self} Attempted to return non-V3 pool {pool_helper}! {pool_address=}, {token_addresses=}, {pool_fee=}"
                    return pool_helper
                else:
                    self._untracked_pools.add(pool_address)
                    raise ManagerError(
                        f"Pool {pool_address} is not associated with this DEX"
                    )

            if self._snapshot:
                v3liquiditypool_kwargs.update(
                    {
                        "tick_bitmap": self._snapshot.get_tick_bitmap(
                            pool_address
                        ),
                        "tick_data": self._snapshot.get_tick_data(
                            pool_address
                        ),
                    }
                )
            else:
                print("SNAPSHOT NOT AVAILABLE!")

            # The pool is unknown, so build and add it
            try:
                pool_helper = V3LiquidityPool(
                    address=pool_address,
                    lens=self._lens,
                    silent=silent,
                    factory_address=self._factory_address,
                    factory_init_hash=self._factory_init_hash,
                    **v3liquiditypool_kwargs,
                    state_block=state_block,
                )
            except Exception as e:
                self._untracked_pools.add(pool_address)
                raise ManagerError(
                    f"Could not build V3 pool {pool_address}: {e}"
                ) from e
            else:
                apply_liquidity_updates(pool_helper)
                self._store_pool(pool_helper)
                assert pool_helper.address in self._tracked_pools
                assert isinstance(
                    pool_helper, V3LiquidityPool
                ), f"{self} Attempted to return non-V3 pool {pool_helper}! {pool_address=}, {token_addresses=}, {pool_fee=}"
                return pool_helper

        if not (pool_address is None) ^ (
            token_addresses is None and pool_fee is None
        ):
            raise ValueError(
                "Insufficient arguments provided. Pass address OR tokens & fee"
            )

        if v3liquiditypool_kwargs is None:
            v3liquiditypool_kwargs = dict()

        if pool_address is not None:
            # print(f"building V3 pool from address")
            if token_addresses is not None or pool_fee is not None:
                raise ValueError(
                    "Conflicting arguments provided. Pass address OR tokens+fee"
                )
            pool_address = to_checksum_address(pool_address)
        elif token_addresses is not None and pool_fee is not None:
            # print(f"building V3 pool from address and fee")
            if len(token_addresses) != 2:
                raise ValueError(
                    f"Expected two tokens, found {len(token_addresses)}"
                )
            try:
                erc20token_helpers: List[Erc20Token] = [
                    self._token_manager.get_erc20token(
                        address=token_address,
                        silent=silent,
                    )
                    for token_address in token_addresses
                ]
            except Exception:
                raise ManagerError("Could not build Erc20Token helper")

            # dictionary key pair is sorted by address
            erc20token_helpers = sorted(erc20token_helpers)
            tokens_key: Tuple[ChecksumAddress, ChecksumAddress] = (
                erc20token_helpers[0].address,
                erc20token_helpers[1].address,
            )

            pool_address = generate_v3_pool_address(
                token_addresses=tokens_key,
                fee=pool_fee,
                factory_address=self._factory_address,
                init_hash=self._factory_init_hash,
            )
        else:
            raise ValueError("THIS BLOCK SHOULD BE UNREACHABLE")

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(
                f"Pool address {pool_address} not associated with factory {self._factory_address}"
            )

        try:
            pool_helper = self._tracked_pools[pool_address]
        except KeyError:
            pool_helper = find_or_build(pool_address, state_block)

        if TYPE_CHECKING:
            assert isinstance(pool_helper, V3LiquidityPool)

        assert isinstance(
            pool_helper, V3LiquidityPool
        ), f"{self} Attempted to return non-V3 pool {pool_helper}! {pool_address=}, {token_addresses=}, {pool_fee=}"
        return pool_helper



# ruff: noqa: F401

from . import exceptions
from . import uniswap
from .arbitrage import (
    ArbitrageCalculationResult,
    FlashBorrowToLpSwap,
    FlashBorrowToLpSwapNew,
    FlashBorrowToLpSwapWithFuture,
    FlashBorrowToRouterSwap,
    UniswapLpCycle,
)
from .chainlink import ChainlinkPriceContract
from .config import get_web3, set_web3
from .fork import AnvilFork
from .functions import next_base_fee
from .logging import logger
from .manager import AllPools, AllTokens, Erc20TokenHelperManager
from .erc20_token import Erc20Token
from .transaction import UniswapTransaction
from .uniswap.managers import (
    UniswapV2LiquidityPoolManager,
    UniswapV3LiquidityPoolManager,
)
from .uniswap.v2_liquidity_pool import CamelotLiquidityPool, LiquidityPool
from .uniswap.v2_multi_liquidity_pool import MultiLiquidityPool
from .uniswap.v3_tick_lens import TickLens
from .uniswap.v3_liquidity_pool import V3LiquidityPool
from .uniswap.v3_dataclasses import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
)
from .uniswap.v3_snapshot import UniswapV3LiquiditySnapshot



from typing import Optional, Dict, Tuple, TYPE_CHECKING
import dataclasses

if TYPE_CHECKING:
    # only necessary for the type hint
    from .v3_liquidity_pool import V3LiquidityPool


@dataclasses.dataclass(slots=True)
class UniswapV3BitmapAtWord:
    bitmap: int = 0
    block: Optional[int] = dataclasses.field(compare=False, default=None)

    def to_dict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass(slots=True)
class UniswapV3LiquidityAtTick:
    liquidityNet: int = 0
    liquidityGross: int = 0
    block: Optional[int] = dataclasses.field(compare=False, default=None)

    def to_dict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass(slots=True)
class UniswapV3LiquidityEvent:
    block_number: int
    liquidity: int
    tick_lower: int
    tick_upper: int
    tx_index: int


@dataclasses.dataclass(slots=True, eq=False)
class UniswapV3PoolExternalUpdate:
    block_number: int = dataclasses.field(compare=False)
    liquidity: Optional[int] = None
    sqrt_price_x96: Optional[int] = None
    tick: Optional[int] = None
    liquidity_change: Optional[
        Tuple[
            int,  # Liquidity
            int,  # TickLower
            int,  # TickUpper
        ]
    ] = None
    tx: Optional[str] = dataclasses.field(compare=False, default=None)


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolState:
    pool: "V3LiquidityPool"
    liquidity: int
    sqrt_price_x96: int
    tick: int
    tick_bitmap: Optional[
        Dict[int, UniswapV3BitmapAtWord]
    ] = dataclasses.field(
        # compare=False,
        default=None,
    )
    tick_data: Optional[
        Dict[int, UniswapV3LiquidityAtTick]
    ] = dataclasses.field(
        # compare=False,
        default=None,
    )


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolSimulationResult:
    amount0_delta: int
    amount1_delta: int
    current_state: UniswapV3PoolState = dataclasses.field(compare=False)
    future_state: UniswapV3PoolState = dataclasses.field(compare=False)



from typing import Optional, Callable
from web3 import Web3

_web3: Optional[Web3] = None


def get_web3() -> Optional[Web3]:
    return _web3


def set_web3(w3: Web3):
    connected_method: Optional[Callable] = None

    for method_name in ("is_connected", "isConnected"):
        try:
            connected_method = getattr(w3, method_name)
        except AttributeError:
            pass
        else:
            break

    if connected_method is None:
        raise ValueError("Provided web3 object has no 'connected' method")

    if connected_method() is False:
        raise ValueError("Web3 object is not connected.")

    global _web3
    _web3 = w3



import os
import socket
import subprocess
from typing import Any, Dict, Optional

import ujson  # type: ignore[import]
from web3 import IPCProvider, Web3


class AnvilFork:
    """
    Launch an Anvil fork via subprocess and provide various methods for
    interacting with it via JSON-RPC over the built-in IPC socket.
    """

    _SOCKET_BUFFER_SIZE = 4096  # https://docs.python.org/3/library/socket.html#socket.socket.recv

    def __init__(
        self,
        fork_url: str,
        hardfork: str = "shanghai",
        gas_limit: int = 30_000_000,
        port: Optional[int] = None,
        fork_block: Optional[int] = None,
        chain_id: Optional[int] = None,
        base_fee: Optional[int] = None,
        ipc_path: Optional[str] = None,
        mnemonic: str = "patient rude simple dog close planet oval animal hunt sketch suspect slim",
    ):
        if not port:
            with socket.socket() as sock:
                sock.bind(("", 0))
                _, port = sock.getsockname()
            self.port = port

        if not ipc_path:
            ipc_path = f"/tmp/anvil-{self.port}.ipc"

        command = []
        command.append("anvil")
        command.append("--silent")
        command.append("--auto-impersonate")
        command.append("--no-rate-limit")
        command.append(f"--fork-url={fork_url}")
        command.append(f"--hardfork={hardfork}")
        command.append(f"--gas-limit={gas_limit}")
        command.append(f"--port={self.port}")
        command.append(f"--ipc={ipc_path}")
        command.append(f"--mnemonic={mnemonic}")
        if fork_block:
            command.append(f"--fork-block-number={fork_block}")
        if chain_id:
            command.append(f"--chain-id={chain_id}")
        if base_fee:
            command.append(f"--base-fee={base_fee}")

        self._process = subprocess.Popen(command)

        self.http_url = f"http://localhost:{self.port}"
        self.ws_url = f"ws://localhost:{self.port}"
        self.ipc_path = ipc_path

        self.w3 = Web3(IPCProvider(ipc_path))

        # Web3.py v5 method is 'isConnected', v6 is 'is_connected'
        for method_name in ("isConnected", "is_connected"):
            if is_connected_method := getattr(self.w3, method_name, None):
                break
        if is_connected_method is None:
            raise ValueError("Web3 provider cannot be tested")
        while is_connected_method() is False:
            continue

        self.block = (
            fork_block if fork_block else self.w3.eth.get_block_number()
        )
        self.base_fee = (
            base_fee
            if base_fee
            else self.w3.eth.get_block(self.block)["baseFeePerGas"]
        )
        self.base_fee_next: Optional[int] = None
        self.socket = socket.socket(socket.AF_UNIX)
        self.socket.connect(self.ipc_path)

    def __del__(self):
        self._process.terminate()
        self._process.wait()
        if os.path.exists(self.ipc_path):
            os.remove(self.ipc_path)

    def create_access_list(self, transaction: Dict) -> list:
        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": 1,
                        "method": "eth_createAccessList",
                        "params": [transaction],
                    }
                ),
                encoding="utf-8",
            ),
        )
        result: dict = ujson.loads(self.socket.recv(self._SOCKET_BUFFER_SIZE))
        if result.get("error"):
            raise Exception(f"Error creating access list! Response: {result}")
        return result["result"]["accessList"]

    def reset(
        self,
        fork_url: Optional[str] = None,
        block_number: Optional[int] = None,
    ) -> None:
        forking_params: Dict[str, Any] = dict()
        if fork_url is not None:
            forking_params["jsonRpcUrl"] = fork_url
        if block_number is not None:
            forking_params["blockNumber"] = block_number

        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": 1,
                        "method": "anvil_reset",
                        "params": [
                            {
                                "forking": forking_params,
                            }
                        ],
                    }
                ),
                encoding="utf-8",
            ),
        )
        result: dict = ujson.loads(self.socket.recv(self._SOCKET_BUFFER_SIZE))
        if result.get("error"):
            raise Exception(f"Error: {result}")
        else:
            if block_number:
                self.block = block_number

    def return_to_snapshot(self, id: str) -> None:
        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": 1,
                        "method": "evm_revert",
                        "params": [id],
                    }
                ),
                encoding="utf-8",
            ),
        )
        result: dict = ujson.loads(self.socket.recv(self._SOCKET_BUFFER_SIZE))
        if result.get("error"):
            raise Exception(
                f"Error reverting to previous snapshot! Response: {result}"
            )

    def set_next_base_fee(self, fee: int) -> None:
        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": 1,
                        "method": "anvil_setNextBlockBaseFeePerGas",
                        "params": [hex(fee)],
                    }
                ),
                encoding="utf-8",
            ),
        )
        result: dict = ujson.loads(self.socket.recv(self._SOCKET_BUFFER_SIZE))
        if result.get("error"):
            raise Exception(
                f"Error setting next block base fee! Response: {result}"
            )
        else:
            self.base_fee_next = fee

    def set_snapshot(self) -> None:
        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": 1,
                        "method": "evm_snapshot",
                        "params": [],
                    }
                ),
                encoding="utf-8",
            ),
        )
        result: dict = ujson.loads(self.socket.recv(self._SOCKET_BUFFER_SIZE))
        if result.get("error"):
            raise Exception(f"Error setting snapshot! Response: {result}")



from typing import List

from ..erc20_token import Erc20Token
from .v2_liquidity_pool import LiquidityPool


class MultiLiquidityPool:
    def __init__(
        self,
        token_in: Erc20Token,
        token_out: Erc20Token,
        pool_addresses: List[str],
        pool_tokens: List[List[Erc20Token]],
        name: str = "",
        update_method: str = "polling",
        silent: bool = False,
    ):
        self.token_in = token_in
        self.token_out = token_out
        self.token_in_quantity = 0
        self.token_out_quantity = 0
        self.init = True

        if len(pool_addresses) != len(pool_tokens):
            raise ValueError(
                "Number of pool addresses and token pairs must match!"
            )

        if not (len(pool_addresses) > 1):
            raise ValueError(
                f"Expected 2 pool addresses, found {len(pool_addresses)}"
            )

        number_of_pools = len(pool_addresses)

        # build the list of pool objects for the given addresses
        self._pools = []
        for i in range(number_of_pools):
            self._pools.append(
                LiquidityPool(
                    address=pool_addresses[i],
                    update_method=update_method,
                    tokens=pool_tokens[i],
                    silent=silent,
                )
            )
        self.pool_addresses = pool_addresses

        if not (
            (token_in == self._pools[0].token0)
            or (token_in == self._pools[0].token1)
        ):
            raise ValueError(
                f"First LP does not contain the submitted token_in ({token_in})"
            )

        if not (
            (token_out == self._pools[-1].token0)
            or (token_out == self._pools[-1].token1)
        ):
            raise ValueError(
                f"Last LP does not contain the submitted token_out ({token_out})"
            )

        # check that pools have a valid token path
        for i in range(number_of_pools - 1):
            if not (
                (self._pools[i].token0 == self._pools[i + 1].token0)
                or (self._pools[i].token1 == self._pools[i + 1].token1)
            ):
                raise ValueError(
                    f"LPs {self._pools[i]} and {self._pools[i+1]} do not share a common token!"
                )

        if name:
            self.name = name

    def update_reserves(
        self,
        silent: bool = False,
        print_reserves: bool = True,
        print_ratios: bool = True,
    ) -> bool:
        """
        Checks each liquidity pool for updates by passing a call to .update_reserves(), which returns False if there are no updates.
        Will calculate arbitrage amounts only after checking all pools and finding an update, or on startup (via the self._init variable)
        """

        recalculate = False

        if self.init is True:
            self.init = False
            recalculate = True

        for pool in self._pools:
            if pool.update_reserves(
                silent=silent,
                print_reserves=print_reserves,
                print_ratios=print_ratios,
            ):
                recalculate = True

        if recalculate:
            self.calculate_multipool_tokens_out_from_tokens_in(
                token_in=self.token_in,
                token_in_quantity=self.token_in_quantity,
                silent=silent,
            )
            return True
        else:
            return False

    def calculate_multipool_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        silent: bool = False,
    ) -> None:
        """
        Calculates the expected token OUTPUT from the last pool for a given token INPUT to the first pool at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        and uses the appropriate formula
        """

        number_of_pools = len(self._pools)

        for i in range(number_of_pools):
            # determine the output token for this pool
            if token_in.address == self._pools[i].token0.address:
                token_out = self._pools[i].token1
            elif token_in.address == self._pools[i].token1.address:
                token_out = self._pools[i].token0
            else:
                print("wtf?")
                raise Exception

            # calculate the swap output from this pool
            token_out_quantity = self._pools[
                i
            ].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
            )

            if i == number_of_pools - 1:
                # if we've reached the last pool, build amounts_out and store the output quantity
                self.token_out_quantity = token_out_quantity
            else:
                # otherwise, use the output as input on the next loop
                token_in = token_out
                token_in_quantity = token_out_quantity

        self._build_multipool_amounts_out(
            token_in=self.token_in,
            token_in_quantity=self.token_in_quantity,
            silent=silent,
        )

    def update_balance(
        self,
        token_in_quantity: int,
        silent: bool = False,
    ):
        self.token_in_quantity = token_in_quantity
        self.calculate_multipool_tokens_out_from_tokens_in(
            token_in=self.token_in,
            token_in_quantity=self.token_in_quantity,
            silent=silent,
        )

    def __str__(self):
        """
        Return the pool name when the object is included in a print statement, or cast as a string
        """
        return self.name

    def _build_multipool_amounts_out(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        silent: bool = True,
    ) -> None:
        number_of_pools = len(self._pools)

        self.pools_amounts_out = []

        for i in range(number_of_pools):
            # determine the output token for pool0
            if token_in.address == self._pools[i].token0.address:
                token_out = self._pools[i].token1
            elif token_in.address == self._pools[i].token1.address:
                token_out = self._pools[i].token0
            else:
                print("wtf?")
                raise Exception

            # calculate the swap output through pool[i]
            token_out_quantity = self._pools[
                i
            ].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
            )

            if not silent:
                print(
                    f"Swap {token_in_quantity} {token_in} for {token_out_quantity} {token_out} via {self._pools[i]}"
                )

            if token_in.address == self._pools[i].token0.address:
                self.pools_amounts_out.append([0, token_out_quantity])
            elif token_in.address == self._pools[i].token1.address:
                self.pools_amounts_out.append([token_out_quantity, 0])

            # use the swap output as the swap input through the next pool
            token_in = token_out
            token_in_quantity = token_out_quantity



from typing import Optional


def next_base_fee(
    parent_base_fee: int,
    parent_gas_used: int,
    parent_gas_limit: int,
    min_base_fee: Optional[int] = None,
    base_fee_max_change_denominator: int = 8,  # limits the maximum base fee increase per block to 1/8 (12.5%)
    elasticity_multiplier: int = 2,
) -> int:
    """
    Calculate next base fee for an EIP-1559 compatible blockchain. The
    formula is taken from the example code in the EIP-1559 proposal (ref:
    https://eips.ethereum.org/EIPS/eip-1559).

    The default values for `base_fee_max_change_denominator` and
    `elasticity_multiplier` are taken from EIP-1559.

    Enforces `min_base_fee` if provided.
    """

    last_gas_target = parent_gas_limit // elasticity_multiplier

    if parent_gas_used == last_gas_target:
        next_base_fee = parent_base_fee
    elif parent_gas_used > last_gas_target:
        gas_used_delta = parent_gas_used - last_gas_target
        base_fee_delta = max(
            parent_base_fee
            * gas_used_delta
            // last_gas_target
            // base_fee_max_change_denominator,
            1,
        )
        next_base_fee = parent_base_fee + base_fee_delta
    else:
        gas_used_delta = last_gas_target - parent_gas_used
        base_fee_delta = (
            parent_base_fee
            * gas_used_delta
            // last_gas_target
            // base_fee_max_change_denominator
        )
        next_base_fee = parent_base_fee - base_fee_delta

    return max(min_base_fee, next_base_fee) if min_base_fee else next_base_fee



# TODO: investigate combining update blocks (is a separate liquidity update block necessary?)
# TODO: compare blocks in bitmaps, ticks, etc.

import dataclasses
import warnings
from bisect import bisect_left
from decimal import Decimal
from threading import Lock
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address

from ..config import get_web3
from ..dex.uniswap import TICKLENS_ADDRESSES
from ..exceptions import (
    BitmapWordUnavailableError,
    BlockUnavailableError,
    EVMRevertError,
    ExternalUpdateError,
    LiquidityPoolError,
    NoPoolStateAvailable,
    ZeroSwapError,
)
from ..logging import logger
from ..manager import AllPools, Erc20TokenHelperManager
from ..erc20_token import Erc20Token
from ..baseclasses import PoolHelper
from .abi import UNISWAP_V3_POOL_ABI
from .v3_functions import generate_v3_pool_address
from .v3_libraries import LiquidityMath, SwapMath, TickBitmap, TickMath
from .v3_libraries.functions import to_int256
from .v3_tick_lens import TickLens
from .v3_dataclasses import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
)


class V3LiquidityPool(PoolHelper):
    __slots__: Tuple[str, ...] = (
        "_extra_words",
        "_fee",
        "_pool_state_archive",
        "_liquidity_lock",
        "_slot0_lock",
        "_sparse_bitmap",
        "_tick_spacing",
        "_update_block",
        "_update_log",
        "_update_method",
        "_w3_contract",
        "_w3",
        "address",
        "factory",
        "lens",
        "liquidity_update_block",
        "name",
        "state",
        "token0",
        "token1",
    )

    # Holds a reference to a TickLens contract object. This is a singleton
    # contract so there is no need to create separate references for each pool.
    # Dict is keyed by a tuple of chain ID and factory address
    _lens_contracts: Dict[Tuple[int, ChecksumAddress], TickLens] = dict()

    uniswap_version = 3

    _TICKSPACING_BY_FEE = {
        100: 1,
        500: 10,
        3000: 60,
        10000: 200,
    }

    def __init__(
        self,
        address: str,
        fee: Optional[int] = None,
        lens: Optional[TickLens] = None,
        tokens: Optional[List[Erc20Token]] = None,
        name: str = "",
        update_method: Optional[str] = None,
        abi: Optional[list] = None,
        factory_address: Optional[str] = None,
        factory_init_hash: Optional[str] = None,
        extra_words: int = 10,
        silent: bool = False,
        tick_data: Optional[dict] = None,
        tick_bitmap: Optional[dict] = None,
        state_block: Optional[int] = None,
    ):
        self.address = to_checksum_address(address)
        abi = abi or UNISWAP_V3_POOL_ABI
        self.state: UniswapV3PoolState = UniswapV3PoolState(
            pool=self,
            liquidity=0,
            sqrt_price_x96=0,
            tick=0,
            tick_bitmap=dict(),
            tick_data=dict(),
        )

        _web3 = get_web3()
        if _web3 is not None:
            self._w3 = _web3
        else:
            from brownie import Contract  # type: ignore[import] # noqa: F401
            from brownie import multicall as brownie_multicall  # noqa: F401
            from brownie import web3 as brownie_web3

            if brownie_web3.isConnected():
                self._w3 = brownie_web3
            else:
                raise ValueError("No connected web3 object provided.")

        self._w3_contract = self._w3.eth.contract(
            address=self.address,
            abi=abi,
        )

        # held by methods that manipulate liquidity data
        self._liquidity_lock = Lock()

        # held by methods that manipulate state data held by slot0
        self._slot0_lock = Lock()

        self._update_block = (
            state_block if state_block else self._w3.eth.get_block_number()
        )

        if factory_address:
            self.factory = to_checksum_address(factory_address)
        else:
            self.factory = to_checksum_address(
                self._w3_contract.functions.factory().call()
            )

        if lens:
            self.lens = lens
        else:
            # Use the singleton TickLens helper if available
            try:
                self.lens = self._lens_contracts[
                    (self._w3.eth.chain_id, self.factory)
                ]
            except KeyError:
                self.lens = TickLens(
                    address=TICKLENS_ADDRESSES[self._w3.eth.chain_id][
                        self.factory
                    ]
                )
                self._lens_contracts[
                    (self._w3.eth.chain_id, self.factory)
                ] = self.lens

        token0_address: ChecksumAddress = to_checksum_address(
            self._w3_contract.functions.token0().call()
        )
        token1_address: ChecksumAddress = to_checksum_address(
            self._w3_contract.functions.token1().call()
        )

        if tokens is not None:
            if len(tokens) != 2:
                raise ValueError(
                    f"Expected exactly two tokens, found {len(tokens)}"
                )

            self.token0 = min(tokens)
            self.token1 = max(tokens)

            if not (
                self.token0 == token0_address and self.token1 == token1_address
            ):
                raise ValueError(
                    "Token addresses do not match tokens recorded at contract"
                )
        else:
            _token_manager = Erc20TokenHelperManager(self._w3.eth.chain_id)
            self.token0 = _token_manager.get_erc20token(
                address=token0_address,
                silent=silent,
            )
            self.token1 = _token_manager.get_erc20token(
                address=token1_address,
                silent=silent,
            )

        self._fee: int = fee or self._w3_contract.functions.fee().call()
        self._tick_spacing = self._TICKSPACING_BY_FEE[self._fee]  # immutable

        if factory_address is not None and factory_init_hash is not None:
            computed_pool_address = generate_v3_pool_address(
                token_addresses=[self.token0.address, self.token1.address],
                fee=self._fee,
                factory_address=factory_address,
                init_hash=factory_init_hash,
            )
            if computed_pool_address != self.address:
                raise ValueError(
                    f"Pool address {self.address} does not match deterministic address {computed_pool_address} from factory"
                )

        if name:
            self.name = name
        else:
            self.name = (
                f"{self.token0}-{self.token1} (V3, {self._fee/10000:.2f}%)"
            )

        if update_method is not None:
            warnings.warn(
                "The `update_method` argument to `V3LiquidityPool()` is unused and otherwise ignored. Remove it to stop seeing this message."
            )
            self._update_method = update_method
        self._extra_words = extra_words

        # default to an empty, sparse bitmap with no tick data
        self._sparse_bitmap = True

        if (tick_bitmap is not None) != (tick_data is not None):
            raise ValueError(
                f"Must provide both tick_bitmap and tick_data! Got {tick_bitmap=}, {tick_data=}"
            )

        if tick_bitmap is not None and tick_data is not None:
            # if a snapshot was provided, assume it is complete
            self._sparse_bitmap = False

        if tick_bitmap is not None:
            # transform dict to UniswapV3BitmapAtWord
            self.tick_bitmap = {
                int(word): (
                    UniswapV3BitmapAtWord(**bitmap_at_word)
                    if not isinstance(
                        bitmap_at_word,
                        UniswapV3BitmapAtWord,
                    )
                    else bitmap_at_word
                )
                for word, bitmap_at_word in tick_bitmap.items()
            }

        if tick_data is not None:
            # transform dict to LiquidityAtTick
            self.tick_data = {
                int(tick): (
                    UniswapV3LiquidityAtTick(**liquidity_at_tick)
                    if not isinstance(
                        liquidity_at_tick,
                        UniswapV3LiquidityAtTick,
                    )
                    else liquidity_at_tick
                )
                for tick, liquidity_at_tick in tick_data.items()
            }

        if tick_bitmap is None and tick_data is None:
            logger.debug(
                f"{self} @ {self.address} updating -> {tick_bitmap=}, {tick_data=}"
            )
            word_position, _ = self._get_tick_bitmap_position(self.tick)

            self._update_tick_data_at_word(
                word_position,
                single_word=True,
                block_number=self._update_block,
            )

        self.liquidity = self._w3_contract.functions.liquidity().call(
            block_identifier=self._update_block
        )

        (
            self.sqrt_price_x96,
            self.tick,
            *_,
        ) = self._w3_contract.functions.slot0().call(
            block_identifier=self._update_block
        )

        self._update_pool_state()

        self._pool_state_archive: Dict[int, UniswapV3PoolState] = {
            self._update_block: self.state
        }

        AllPools(self._w3.eth.chain_id)[self.address] = self

        if not silent:
            logger.info(self.name)
            logger.info(f"• Token 0: {self.token0}")
            logger.info(f"• Token 1: {self.token1}")
            logger.info(f"• Fee: {self._fee}")
            logger.info(f"• Liquidity: {self.liquidity}")
            logger.info(f"• SqrtPrice: {self.sqrt_price_x96}")
            logger.info(f"• Tick: {self.tick}")

    def __eq__(self, other) -> bool:
        if issubclass(type(other), PoolHelper):
            return self.address == other.address
        elif isinstance(other, str):
            return self.address.lower() == other.lower()
        else:
            raise NotImplementedError

    def __getstate__(self) -> dict:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        keys_to_remove = (
            "_liquidity_lock",
            "_slot0_lock",
            "_w3",
            "_w3_contract",
            "lens",
        )

        try:
            self.__slots__
        except AttributeError:
            pass
        else:
            return {
                attr_name: getattr(self, attr_name, None)
                for attr_name in self.__slots__
                if attr_name not in keys_to_remove
            }

        return {
            key: value
            for key, value in self.__dict__.items()
            if key not in keys_to_remove
        }

    def __hash__(self):
        return hash(self.address)

    def __repr__(self):
        return f"V3LiquidityPool(address={self.address}, token0={self.token0}, token1={self.token1}, fee={self._fee})"

    def __setstate__(self, state: dict):
        for key, value in state.items():
            setattr(self, key, value)

    def __str__(self):
        """
        Return the pool name when the object is included in a print statement, or cast as a string
        """
        return self.name

    def _get_tick_bitmap_position(self, tick) -> Tuple[int, int]:
        """
        Retrieves the wordPosition and bitPosition for the input tick

        This function corrects internally for tick spacing! e.g. tick=600 is the
        11th initialized tick for an LP with tickSpacing of 60, starting at 0.
        Each "word" in the tickBitmap holds 256 initialized positions, so the 11th
        position of the 1st word will represent tick=600.

        Calling `get_tick_bitmap_position(600)` returns (0,10), where:
            0 = wordPosition (zero-indexed)
            10 = bitPosition (zero-indexed)
        """
        return TickBitmap.position(int(Decimal(tick) // self._tick_spacing))

    def _update_pool_state(self) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=self.liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=self.tick_bitmap.copy(),
            tick_data=self.tick_data.copy(),
        )

    def _update_tick_data_at_word(
        self,
        word_position: int,
        single_word: bool = False,
        block_number: Optional[int] = None,
    ) -> None:
        """
        Update the initialized tick values at a specific word (a 32 byte number
        representing 256 ticks at the tickSpacing interval). Store
        the liquidity values in the `self.tick_data` dictionary using the tick
        as the key, and update the `self.tick_bitmap` dictionary.

        This function attempts to use Brownie's built-in multicall for any network
        with the 'multicall2' key set. If available, it will request extra words
        in the direction of the requested position to fill in gaps in `self.tick_data`

        If multicall is set but `self.tick_data` is empty, it will fall back to fetching
        a single word only. This is a time-saving technique since this should only occur
        inside the constructor when the pool helper is being created.

        Uses a lock to guard state-modifying methods that might cause race conditions
        when used with threads.
        """

        # Return immediately if requested word is already known.
        # This can occur in threaded bots. The lock prevents race conditions,
        # but this method might be running simultaneously across different threads.
        # Used to throw an exception, now just returns early.

        if word_position in self.tick_bitmap:
            logger.debug(f"returning early, {word_position=} found")
            logger.debug(self.tick_bitmap[word_position])
            return

        if block_number is None:
            block_number = self._w3.eth.get_block_number()

        with self._liquidity_lock:
            if False:
                pass

            # TODO: re-implement multicall without Brownie
            # # fetch multiple words if multicall is available for the connected network
            # # and single_word mode is not active
            # if (
            #     network.main.CONFIG.active_network.get("multicall2")
            #     and not single_word
            # ):

            #     # limit word values to int16 range
            #     words = set(
            #         range(
            #             max(MIN_INT16, word_position - self.extra_words // 2),
            #             min(MAX_INT16, word_position + self.extra_words // 2),
            #         )
            #     ) - set(self.tick_bitmap)

            #     logger.debug(f"fetching words: {words}")

            #     # fetch the tick bitmaps for the range
            #     try:
            #         with brownie_multicall(block_identifier=block_number):
            #             multicall_tick_bitmaps = {
            #                 word: UniswapV3BitmapAtWord(
            #                     bitmap=self._brownie_contract.tickBitmap(word),
            #                     block=block_number,
            #                 )
            #                 for word in words
            #             }
            #         with brownie_multicall(block_identifier=block_number):
            #             multicall_tick_data = {
            #                 tick: UniswapV3LiquidityAtTick(
            #                     liquidityNet=liquidity_net,
            #                     liquidityGross=liquidity_gross,
            #                     block=block_number,
            #                 )
            #                 for word_position, bitmap in multicall_tick_bitmaps.items()
            #                 for tick, liquidity_net, liquidity_gross in self.lens._brownie_contract.getPopulatedTicksInWord(
            #                     self.address,
            #                     word_position,
            #                 )
            #                 if bitmap
            #             }
            #     except Exception as e:
            #         print(
            #             f"(V3LiquidityPool (_update_tick_data_at_word) (multicall): {e}"
            #         )
            #         print(type(e))
            #         raise
            #     else:
            #         self.tick_bitmap.update(multicall_tick_bitmaps)
            #         self.tick_data.update(multicall_tick_data)
            #         self.liquidity_update_block = block_number

            # fetch words one by one (single_tick = True)
            else:
                try:
                    if (
                        single_tick_bitmap
                        := self._w3_contract.functions.tickBitmap(
                            word_position
                        ).call(
                            block_identifier=block_number,
                        )
                    ):
                        single_tick_data = self.lens._w3_contract.functions.getPopulatedTicksInWord(
                            self.address, word_position
                        ).call(
                            block_identifier=block_number,
                        )
                except Exception as e:
                    print(
                        f"(V3LiquidityPool) (_update_tick_data_at_word) (single tick): {e}"
                    )
                    print(type(e))
                    raise
                else:
                    self.tick_bitmap[word_position] = UniswapV3BitmapAtWord(
                        bitmap=single_tick_bitmap,
                        block=block_number,
                    )
                    if single_tick_bitmap:
                        for (
                            tick,
                            liquidity_net,
                            liquidity_gross,
                        ) in single_tick_data:
                            self.tick_data[tick] = UniswapV3LiquidityAtTick(
                                liquidityNet=liquidity_net,
                                liquidityGross=liquidity_gross,
                                block=block_number,
                            )

    def _uniswap_v3_pool_swap(
        self,
        zeroForOne: bool,
        amount_specified: int,
        sqrt_price_limit_x96: int,
        override_start_liquidity: Optional[int] = None,
        override_start_sqrt_price_x96: Optional[int] = None,
        override_start_tick: Optional[int] = None,
        override_tick_data: Optional[dict] = None,
        override_tick_bitmap: Optional[dict] = None,
    ) -> Tuple[int, int, int, int, int]:
        """
        This function is ported and adapted from the UniswapV3Pool.sol contract
        at https://github.com/Uniswap/v3-core/blob/main/contracts/UniswapV3Pool.sol

        It is called by the `calculate_tokens_in_from_tokens_out` and `calculate_tokens_out_from_tokens_in` methods to calculate
        swap amounts, ticks crossed, liquidity changes at various ticks, etc.
        """

        @dataclasses.dataclass(slots=True, eq=False)
        class SwapCache:
            liquidityStart: int
            tickCumulative: int

        @dataclasses.dataclass(slots=True, eq=False)
        class SwapState:
            amountSpecifiedRemaining: int
            amountCalculated: int
            sqrtPriceX96: int
            tick: int
            liquidity: int

        @dataclasses.dataclass(slots=True, eq=False)
        class StepComputations:
            sqrtPriceStartX96: int = 0
            tickNext: int = 0
            initialized: bool = False
            sqrtPriceNextX96: int = 0
            amountIn: int = 0
            amountOut: int = 0
            feeAmount: int = 0

        if amount_specified == 0:
            raise EVMRevertError("AS")

        if override_start_liquidity is not None:
            liquidity = override_start_liquidity
        else:
            liquidity = self.liquidity

        if override_start_sqrt_price_x96 is not None:
            sqrt_price_x96 = override_start_sqrt_price_x96
        else:
            sqrt_price_x96 = self.sqrt_price_x96

        if override_start_tick is not None:
            tick = override_start_tick
        else:
            tick = self.tick

        if override_tick_bitmap is not None:
            _tick_bitmap = override_tick_bitmap
        else:
            _tick_bitmap = self.tick_bitmap.copy()

        if override_tick_data is not None:
            _tick_data = override_tick_data
        else:
            _tick_data = self.tick_data.copy()

        if not (
            sqrt_price_limit_x96 < sqrt_price_x96
            and sqrt_price_limit_x96 > TickMath.MIN_SQRT_RATIO
            if zeroForOne
            else sqrt_price_limit_x96 > sqrt_price_x96
            and sqrt_price_limit_x96 < TickMath.MAX_SQRT_RATIO
        ):
            raise EVMRevertError("SPL")

        cache = SwapCache(
            liquidityStart=liquidity,
            tickCumulative=0,
            # ignored attributes:
            #   - blockTimestamp
            #   - feeProtocol
            #   - secondsPerLiquidityCumulativeX128
            #   - computedLatestObservation
        )

        exactInput: bool = amount_specified > 0

        state = SwapState(
            amountSpecifiedRemaining=amount_specified,
            amountCalculated=0,
            sqrtPriceX96=sqrt_price_x96,
            tick=tick,
            liquidity=cache.liquidityStart,
            # ignored attributes:
            #   - feeGrowthGlobalX128
            #   - protocolFee
        )

        while (
            state.amountSpecifiedRemaining != 0
            and state.sqrtPriceX96 != sqrt_price_limit_x96
        ):
            step = StepComputations()

            step.sqrtPriceStartX96 = state.sqrtPriceX96

            while True:
                try:
                    (
                        step.tickNext,
                        step.initialized,
                    ) = TickBitmap.nextInitializedTickWithinOneWord(
                        _tick_bitmap,
                        state.tick,
                        self._tick_spacing,
                        zeroForOne,
                    )
                except BitmapWordUnavailableError as e:
                    missing_word = e.args[1]
                    if self._sparse_bitmap:
                        logger.debug(
                            f"(swap) {self.name} fetching word {missing_word}"
                        )
                        self._update_tick_data_at_word(missing_word)
                    else:
                        # bitmap is complete, so mark the word as empty
                        # self.tick_bitmap[missing_word] = UniswapV3BitmapAtWord()
                        _tick_bitmap[missing_word] = UniswapV3BitmapAtWord()
                else:
                    # nextInitializedTickWithinOneWord will search up to 256 ticks away, which may
                    # return a tick in an adjacent word if there are no initialized ticks in the current word.
                    # This word may not be known to the helper, so check and fetch the containing word for this tick

                    # BUGFIX: previously called position directly, which implies tickSpacing=1,
                    # so the call returned an inaccurate word and short-circuited the optimization
                    tick_next_word, _ = self._get_tick_bitmap_position(
                        step.tickNext
                    )

                    if (
                        self._sparse_bitmap
                        and tick_next_word not in _tick_bitmap
                    ):
                        logger.debug(
                            f"tickNext={step.tickNext} out of range! Fetching word={tick_next_word}"
                            f"\n{self.name}"
                        )
                        self._update_tick_data_at_word(
                            tick_next_word,
                            single_word=True,
                        )
                    break

            # ensure that we do not overshoot the min/max tick, as the tick bitmap is not aware of these bounds
            if step.tickNext < TickMath.MIN_TICK:
                step.tickNext = TickMath.MIN_TICK
            elif step.tickNext > TickMath.MAX_TICK:
                step.tickNext = TickMath.MAX_TICK

            # get the price for the next tick
            step.sqrtPriceNextX96 = TickMath.getSqrtRatioAtTick(step.tickNext)

            # compute values to swap to the target tick, price limit, or point where input/output amount is exhausted
            (
                state.sqrtPriceX96,
                step.amountIn,
                step.amountOut,
                step.feeAmount,
            ) = SwapMath.computeSwapStep(
                state.sqrtPriceX96,
                sqrt_price_limit_x96
                if (
                    step.sqrtPriceNextX96 < sqrt_price_limit_x96
                    if zeroForOne
                    else step.sqrtPriceNextX96 > sqrt_price_limit_x96
                )
                else step.sqrtPriceNextX96,
                state.liquidity,
                state.amountSpecifiedRemaining,
                self._fee,
            )

            if exactInput:
                state.amountSpecifiedRemaining -= to_int256(
                    step.amountIn + step.feeAmount
                )
                state.amountCalculated = to_int256(
                    state.amountCalculated - step.amountOut
                )
            else:
                state.amountSpecifiedRemaining += to_int256(step.amountOut)
                state.amountCalculated = to_int256(
                    state.amountCalculated + step.amountIn + step.feeAmount
                )

            # shift tick if we reached the next price
            if state.sqrtPriceX96 == step.sqrtPriceNextX96:
                # if the tick is initialized, run the tick transition
                if step.initialized:
                    tick_next = step.tickNext
                    liquidityNet = _tick_data[tick_next].liquidityNet

                    if zeroForOne:
                        liquidityNet = -liquidityNet

                    state.liquidity = LiquidityMath.addDelta(
                        state.liquidity, liquidityNet
                    )

                state.tick = step.tickNext - 1 if zeroForOne else step.tickNext

            elif state.sqrtPriceX96 != step.sqrtPriceStartX96:
                # recompute unless we're on a lower tick boundary (i.e. already transitioned ticks), and haven't moved
                state.tick = TickMath.getTickAtSqrtRatio(state.sqrtPriceX96)

        amount0, amount1 = (
            (
                amount_specified - state.amountSpecifiedRemaining,
                state.amountCalculated,
            )
            if zeroForOne == exactInput
            else (
                state.amountCalculated,
                amount_specified - state.amountSpecifiedRemaining,
            )
        )

        return (
            amount0,
            amount1,
            state.sqrtPriceX96,
            state.liquidity,
            state.tick,
        )

    @property
    def liquidity(self) -> int:
        return self.state.liquidity

    @liquidity.setter
    def liquidity(self, new_liquidity: int) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=new_liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=self.tick_bitmap,
            tick_data=self.tick_data,
        )

    @property
    def sqrt_price_x96(self) -> int:
        return self.state.sqrt_price_x96

    @sqrt_price_x96.setter
    def sqrt_price_x96(self, new_sqrt_price_x96: int) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=self.liquidity,
            sqrt_price_x96=new_sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=self.tick_bitmap,
            tick_data=self.tick_data,
        )

    @property
    def tick(self) -> int:
        return self.state.tick

    @tick.setter
    def tick(self, new_tick: int) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=self.liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=new_tick,
            tick_bitmap=self.tick_bitmap,
            tick_data=self.tick_data,
        )

    @property
    def tick_bitmap(self) -> Dict[int, UniswapV3BitmapAtWord]:
        if TYPE_CHECKING:
            assert self.state.tick_bitmap is not None
        return self.state.tick_bitmap

    @tick_bitmap.setter
    def tick_bitmap(self, new_tick_bitmap: dict) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=self.liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=new_tick_bitmap,
            tick_data=self.tick_data,
        )

    @property
    def tick_data(self) -> Dict[int, UniswapV3LiquidityAtTick]:
        if TYPE_CHECKING:
            assert self.state.tick_data is not None
        return self.state.tick_data

    @tick_data.setter
    def tick_data(self, new_tick_data: dict) -> None:
        self.state = UniswapV3PoolState(
            pool=self,
            liquidity=self.liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=self.tick_bitmap,
            tick_data=new_tick_data,
        )

    def auto_update(
        self,
        silent: bool = True,
        block_number: Optional[int] = None,
    ) -> Tuple[bool, UniswapV3PoolState]:
        """
        Retrieves the current slot0 and liquidity values from the LP, stores any that have changed,
        and returns a tuple with a status boolean indicating whether any update was found,
        and a dictionary holding current state values:
            - liquidity
            - sqrt_price_x96
            - tick

        Uses a lock to guard state-modifying methods that might cause race conditions
        when used with threads.
        """

        with self._slot0_lock:
            updated = False

            # use the block_number if provided, otherwise pull from web3
            if block_number is None:
                block_number = self._w3.eth.get_block_number()

            if block_number < self._update_block:
                raise ExternalUpdateError(
                    f"Current state recorded at block {self._update_block}, received update for stale block {block_number}"
                )

            (
                _sqrt_price_x96,
                _tick,
                *_,
            ) = self._w3_contract.functions.slot0().call(
                block_identifier=block_number,
            )
            _liquidity = self._w3_contract.functions.liquidity().call(
                block_identifier=block_number
            )

            if self.sqrt_price_x96 != _sqrt_price_x96:
                updated = True
                self.sqrt_price_x96 = _sqrt_price_x96

            if self.tick != _tick:
                updated = True
                self.tick = _tick

            if self.liquidity != _liquidity:
                updated = True
                self.liquidity = _liquidity

            if updated:
                self._update_pool_state()
                # WIP: maintain a dict of pool states by block to unwind updates that were removed by a re-org
                self._pool_state_archive[block_number] = self.state

            if not silent:
                logger.info(f"Liquidity: {self.liquidity}")
                logger.info(f"SqrtPriceX96: {self.sqrt_price_x96}")
                logger.info(f"Tick: {self.tick}")

            # WORKAROUND: update the block even if there are no state changes
            # pools were being repeatedly caught by "stale pool" checks
            self._update_block = block_number

        return updated, self.state

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        override_state: Optional[UniswapV3PoolState] = None,
    ) -> int:
        """
        This function implements the common degenbot interface `calculate_tokens_out_from_tokens_in`
        to calculate the number of tokens withdrawn (out) for a given number of tokens deposited (in).

        It is similar to calling quoteExactInputSingle using the quoter contract with arguments:
        `quoteExactInputSingle(
            tokenIn=token_in,
            tokenOut=[automatically determined by helper],
            fee=[automatically determined by helper],
            amountIn=token_in_quantity,
            sqrt_price_limitX96 = 0
        )` which returns the value `amountOut`

        Note that this wrapper function always assumes that the sqrt_price_limitx96 argument is unset,
        thus the swap calculation will continue until the target amount is satisfied, regardless of
        price impact.

        Accepts a dictionary of state values (`override_state`) to allow calculations beginning from an
        arbitrary starting point. This dictionary must have one or more of the following keys:
            - 'liquidity'
            - 'sqrt_price_x96'
            - 'tick'
            - 'tick_data'  (not yet implemented)
            - 'tick_bitmap'  (not yet implemented)
        """

        if token_in not in (self.token0, self.token1):
            raise ValueError("token_in not found!")

        if override_state:
            logger.debug(f"V3 calc with overridden state: {override_state}")

        # determine whether the swap is token0 -> token1
        zeroForOne = True if token_in == self.token0 else False

        try:
            amount0_delta, amount1_delta, *_ = self._uniswap_v3_pool_swap(
                zeroForOne=zeroForOne,
                amount_specified=token_in_quantity,
                sqrt_price_limit_x96=(
                    TickMath.MIN_SQRT_RATIO + 1
                    if zeroForOne
                    else TickMath.MAX_SQRT_RATIO - 1
                ),
                override_start_liquidity=override_state.liquidity
                if override_state
                else None,
                override_start_sqrt_price_x96=override_state.sqrt_price_x96
                if override_state
                else None,
                override_start_tick=override_state.tick
                if override_state
                else None,
                override_tick_bitmap=override_state.tick_bitmap
                if override_state
                else None,
                override_tick_data=override_state.tick_data
                if override_state
                else None,
            )
        except EVMRevertError as e:
            raise LiquidityPoolError(
                f"Simulated execution reverted: {e}"
            ) from e
        else:
            return -amount1_delta if zeroForOne else -amount0_delta

    def calculate_tokens_in_from_tokens_out(
        self,
        token_out: Erc20Token,
        token_out_quantity: int,
        override_state: Optional[UniswapV3PoolState] = None,
    ) -> int:
        """
        This function implements the common degenbot interface `calculate_tokens_in_from_tokens_out`
        to calculate the number of tokens deposited (in) for a given number of tokens withdrawn (out).

        It is similar to calling quoteExactOutputSingle using the quoter contract with arguments:
        `quoteExactOutputSingle(
            tokenIn=[automatically determined by helper],
            tokenOut=token_out,
            fee=[automatically determined by helper],
            amountOut=token_out_quantity,
            sqrt_price_limitX96 = 0
        )` which returns the value `amountIn`

        Note that this wrapper function always assumes that the sqrt_price_limitx96 argument is unset, thus the
        swap calculation will continue until the target amount is satisfied, regardless of price impact

        Accepts a dictionary of state values (`override_state`) to allow calculations beginning from an
        arbitrary starting point. This dictionary must have one or more of the following keys:
            - 'liquidity'
            - 'sqrt_price_x96'
            - 'tick'
            - 'tick_data'  (not yet implemented)
            - 'tick_bitmap'  (not yet implemented)
        """

        if token_out not in (self.token0, self.token1):
            raise ValueError("token_in not found!")

        # determine whether the swap is token0 -> token1
        zeroForOne = True if token_out == self.token1 else False

        try:
            amount0_delta, amount1_delta, *_ = self._uniswap_v3_pool_swap(
                zeroForOne=zeroForOne,
                amount_specified=-token_out_quantity,
                sqrt_price_limit_x96=(
                    TickMath.MIN_SQRT_RATIO + 1
                    if zeroForOne
                    else TickMath.MAX_SQRT_RATIO - 1
                ),
                override_start_liquidity=override_state.liquidity
                if override_state
                else None,
                override_start_sqrt_price_x96=override_state.sqrt_price_x96
                if override_state
                else None,
                override_start_tick=override_state.tick
                if override_state
                else None,
                override_tick_bitmap=override_state.tick_bitmap
                if override_state
                else None,
                override_tick_data=override_state.tick_data
                if override_state
                else None,
            )
        except EVMRevertError as e:
            raise LiquidityPoolError(
                f"Simulated execution reverted: {e}"
            ) from e
        else:
            amountIn, amountOutReceived = (
                (amount0_delta, -amount1_delta)
                if zeroForOne
                else (amount1_delta, -amount0_delta)
            )

            return amountIn

    def external_update(
        self,
        update: Optional[UniswapV3PoolExternalUpdate] = None,
        updates: Optional[dict] = None,
        block_number: Optional[int] = None,
        silent: bool = True,
        fetch_missing: Optional[bool] = None,
        force: bool = False,  # added primarily to support liquidity bootstrapping without excessive refactoring
    ) -> bool:
        """
        Process a `UniswapV3PoolExternalUpdate` with one or more of the following update types:
            - `tick`: int
            - `liquidity`: int
            - `sqrt_price_x96`: int
            - `liquidity_change`: tuple of (liquidity_delta, lower_tick, upper_tick). The delta can be positive or negative to indicate added or removed liquidity.

        `block_number` is validated against the most recently recorded block prior to recording any changes. If `force=True`, the block check is skipped.

        If any update is processed, `self.state` and `self.update_block` are updated.

        Returns a bool indicating whether any updated state value was recorded.

        @dev This method uses a lock to guard state-modifying methods that might cause race conditions when used with threads.
        """

        def is_valid_update_block(block_number) -> bool:
            """
            Check if `block_number` is valid (matches or exceeds the last update block)

            The `force` argument to `external_update` will trigger this to always return True
            """
            return True if force else block_number >= self._update_block

        if fetch_missing is not None:
            raise DeprecationWarning(
                "The fetch_missing argument has been deprecated, to address this exception remove it from any calls to external_update"
            )

        # warnings.warn(
        #     "\n"
        #     + "The `updates` dict argument is deprecated and will be "
        #     + "removed in the future. It has been converted in-place to a "
        #     + "`UniswapV3PoolExternalUpdate` object. Pass this using the "
        #     + "`update=` argument to remove this warning."
        #     + "\n"
        #     + "For the values you've provided, pass this data using "
        #     + "the format: \n"
        #     + "update=UniswapV3PoolExternalUpdate(\n"
        #     + (
        #         f"    liquidity={val}\n"
        #         if (val := updates.get("liquidity")) is not None
        #         else ""
        #     )
        #     + (
        #         f"    sqrt_price_x96={val}\n"
        #         if (val := updates.get("sqrt_price_x96")) is not None
        #         else ""
        #     )
        #     + (
        #         f"    tick={val}\n"
        #         if (val := updates.get("tick")) is not None
        #         else ""
        #     )
        #     + ")",
        # )

        if TYPE_CHECKING:
            assert isinstance(update, UniswapV3PoolExternalUpdate)

        # If a block number was not provided, pull from web3
        if block_number is None:
            if update is not None:
                block_number = update.block_number
            else:
                block_number = self._w3.eth.get_block_number()
                warnings.warn(
                    f"(V3LiquidityPool.external_update) block_number was not provided, using {block_number} from chain"
                )

        if not is_valid_update_block(block_number):
            raise ExternalUpdateError(
                f"Rejected update for block {block_number} in the past, current update block is {self._update_block}"
            )

        if updates and not update:
            update = UniswapV3PoolExternalUpdate(
                block_number=block_number,
                liquidity=updates.get("liquidity"),
                sqrt_price_x96=updates.get("sqrt_price_x96"),
                tick=updates.get("tick"),
                liquidity_change=updates.get("liquidity_change"),
            )

        with self._slot0_lock:
            updated_state = False

            for update_type in ("tick", "liquidity", "sqrt_price_x96"):
                if (
                    update_value := getattr(update, update_type, None)
                ) and update_value != getattr(self, update_type):
                    setattr(self, update_type, update_value)
                    updated_state = True

            if update.liquidity_change:
                (
                    liquidity_delta,
                    lower_tick,
                    upper_tick,
                ) = update.liquidity_change

                if liquidity_delta:
                    updated_state = True

                # adjust in-range liquidity if current tick is within the position's range
                if lower_tick <= self.tick < upper_tick and not force:
                    liquidity_before = self.liquidity
                    self.liquidity += liquidity_delta
                    logger.debug(
                        f"Adjusting in-range liquidity, TX {update.tx}, tick range = [{lower_tick},{upper_tick}], current tick = {self.tick}, {self.address=}, previous liquidity = {liquidity_before}, liquidity change = {liquidity_delta}, current liquidity = {self.liquidity}"
                    )
                    assert (
                        self.liquidity >= 0
                    ), f"{self.address=}, {liquidity_before=}, {self.liquidity=}, {update=} {block_number=} {self.tick=}"

                for i, tick in enumerate([lower_tick, upper_tick]):
                    tick_word, _ = self._get_tick_bitmap_position(tick)

                    if tick_word not in self.tick_bitmap:
                        # the tick bitmap must be available for the word prior to flipping
                        # the initialized status of any tick

                        if self._sparse_bitmap:
                            logger.debug(
                                f"(external_update) {tick_word=} not found in tick_bitmap {self.tick_bitmap.keys()=}"
                            )
                            try:
                                # fetch the single word
                                self._update_tick_data_at_word(
                                    word_position=tick_word,
                                    single_word=True,
                                    # Fetch the word using the previous block as a known "good" state snapshot
                                    block_number=block_number - 1,
                                )
                            except ValueError as e:
                                raise BlockUnavailableError(
                                    f"Could not query chain at block {block_number - 1}"
                                ) from e
                        else:
                            # The bitmap is complete (sparse=False), so mark this word as empty
                            self.tick_bitmap[
                                tick_word
                            ] = UniswapV3BitmapAtWord()

                    # Get the liquidity info for this tick
                    try:
                        tick_liquidity_net, tick_liquidity_gross = (
                            self.tick_data[tick].liquidityNet,
                            self.tick_data[tick].liquidityGross,
                        )
                    except (KeyError, AttributeError):
                        # if it doesn't exist, initialize the tick and set the current values to zero
                        tick_liquidity_net = 0
                        tick_liquidity_gross = 0
                        TickBitmap.flipTick(
                            self.tick_bitmap,
                            tick,
                            self._tick_spacing,
                            update_block=block_number,
                        )

                    # MINT: add liquidity at lower tick (i==0), subtract at upper tick (i==1)
                    # BURN: subtract liquidity at lower tick (i==0), add at upper tick (i==1)
                    # Same equation, but for BURN events the liquidity_delta value is negative
                    new_liquidity_net = (
                        tick_liquidity_net + liquidity_delta
                        if i == 0
                        else tick_liquidity_net - liquidity_delta
                    )
                    new_liquidity_gross = (
                        tick_liquidity_gross + liquidity_delta
                    )

                    # Delete entirely if there is no liquidity referencing this tick, then flip it in the bitmap
                    if new_liquidity_gross == 0:
                        del self.tick_data[tick]
                        TickBitmap.flipTick(
                            self.tick_bitmap,
                            tick,
                            self._tick_spacing,
                            update_block=block_number,
                        )
                    # otherwise record the new values
                    else:
                        self.tick_data[tick] = UniswapV3LiquidityAtTick(
                            liquidityNet=new_liquidity_net,
                            liquidityGross=new_liquidity_gross,
                            block=block_number,
                        )

            if not silent:
                logger.debug(f"Liquidity: {self.liquidity}")
                logger.debug(f"SqrtPriceX96: {self.sqrt_price_x96}")
                logger.debug(f"Tick: {self.tick}")
                logger.debug(
                    f"liquidity event: {liquidity_delta} in tick range [{lower_tick},{upper_tick}], pool: {self.name}"
                    "\n"
                    f"old liquidity: {tick_liquidity_net} net, {tick_liquidity_gross} gross"
                    "\n"
                    f"new liquidity: {new_liquidity_net} net, {new_liquidity_gross} gross"
                )
                logger.debug(
                    f"update block: {block_number} (last={self._update_block})"
                )

            if updated_state:
                self._update_pool_state()
                self._pool_state_archive[block_number] = self.state

                if not force:
                    self._update_block = block_number

            return updated_state

    def restore_state_before_block(
        self,
        block: int,
    ) -> None:
        """
        Restore the last pool state recorded prior to a target block.

        Use this method to maintain consistent state data following a chain
        re-organization.
        """

        # Find the index for the most recent pool state PRIOR to the requested
        # block number.
        #
        # e.g. Calling restore_state_before_block(block=104) for a pool with
        # states at blocks 100, 101, 102, 103, 104. `bisect_left()` returns
        # block_index=3, since block 104 is at index=4. The state held at
        # index=3 is for block 103.
        # block_index = self._pool_state_archive.bisect_left(block)

        known_blocks = list(self._pool_state_archive.keys())
        block_index = bisect_left(known_blocks, block)

        if block_index == 0:
            raise NoPoolStateAvailable(
                f"No pool state known prior to block {block}"
            )

        # The last known state already meets the criterion, so return early
        if block_index == len(known_blocks):
            return

        # Remove states at and after the specified block
        for block in known_blocks[block_index:]:
            del self._pool_state_archive[block]

        restored_block, restored_state = list(
            self._pool_state_archive.items()
        )[-1]

        self.state = restored_state
        self._update_block = restored_block

    def simulate_swap(
        self,
        token_in: Optional[Erc20Token] = None,
        token_in_quantity: Optional[int] = None,
        token_out: Optional[Erc20Token] = None,
        token_out_quantity: Optional[int] = None,
        sqrt_price_limit: Optional[int] = None,
        override_state: Optional[UniswapV3PoolState] = None,
    ) -> UniswapV3PoolSimulationResult:
        """
        [TBD]
        """

        if token_in is None and token_out is None:
            raise ValueError("token_in or token_out not provided")

        if token_in_quantity and token_out_quantity:
            raise ValueError(
                "Provide token_in_quantity or token_out_quantity, not both"
            )

        if token_in is not None:
            if token_in_quantity is None:
                raise ValueError("token_in_quantity not provided")
            if token_in not in (self.token0, self.token1):
                raise ValueError("token_in not provided")

        if token_out is not None:
            if token_out_quantity is None:
                raise ValueError("token_out_quantity not provided")
            if token_out not in (self.token0, self.token1):
                raise ValueError("token_out not provided")

        if 0 in (token_in_quantity, token_out_quantity):
            raise ZeroSwapError("Zero input/output swap requested")

        # determine whether the swap is token0 -> token1
        if token_in is not None:
            zeroForOne = True if token_in == self.token0 else False
        elif token_out is not None:
            zeroForOne = True if token_out == self.token1 else False

        _sqrt_price_limit = (
            sqrt_price_limit
            if sqrt_price_limit is not None
            else (
                TickMath.MIN_SQRT_RATIO + 1
                if zeroForOne
                else TickMath.MAX_SQRT_RATIO - 1
            )
        )

        if token_in_quantity is not None:
            _amount_specified = token_in_quantity
        elif token_out_quantity is not None:
            _amount_specified = -token_out_quantity

        try:
            (
                amount0_delta,
                amount1_delta,
                end_sqrtprice,
                end_liquidity,
                end_tick,
            ) = self._uniswap_v3_pool_swap(
                zeroForOne=zeroForOne,
                amount_specified=_amount_specified,
                sqrt_price_limit_x96=_sqrt_price_limit,
                override_start_liquidity=override_state.liquidity
                if override_state
                else None,
                override_start_sqrt_price_x96=override_state.sqrt_price_x96
                if override_state
                else None,
                override_start_tick=override_state.tick
                if override_state
                else None,
                override_tick_bitmap=override_state.tick_bitmap
                if override_state
                else None,
                override_tick_data=override_state.tick_data
                if override_state
                else None,
            )
        except EVMRevertError as e:
            raise LiquidityPoolError(
                f"Simulated execution reverted: {e}"
            ) from e
        else:
            return UniswapV3PoolSimulationResult(
                amount0_delta=amount0_delta,
                amount1_delta=amount1_delta,
                current_state=self.state,
                future_state=UniswapV3PoolState(
                    pool=self,
                    liquidity=end_liquidity,
                    sqrt_price_x96=end_sqrtprice,
                    tick=end_tick,
                ),
            )
